"""
File Purpose: arithmetic for vectors or components.

[EFF] the idea for efficiency in vector methods
    is that it's more efficient to get the full vector instead of one component at a time,
    if you'll be using more than one component of it.
    For example, if you use electron momentum equation to solve for electric field,
        then 'E' is complicated to calculate, and involves some scalar factors.
        If you get 'E_x' then 'E_y', you'll need to recalculate those factors each time.
        But if you get 'E_xy', then you only need to calculate those factors once
            (assuming that get_E is written appropriately, with this in mind).

[TODO] compatibility with rotations (when implementing rotations):
    infer X,Y,Z from 'component' dimension of array, instead of imposing XYZ.
    E.g. if using a rotated coordinate system xprime, yprime, zprime,
        then infer X,Y,Z as xprime,yprime,zprime. And crash if mixing coordinate systems...?
"""
import numpy as np
import xarray as xr

from ..quantity_loader import QuantityLoader
from ...dimensions import XYZ, YZ_FROM_X, XHAT, YHAT, ZHAT
from ...errors import QuantCalcError
from ...tools import (
    UNSET,
    alias,
    take_along_dimension, join_along_dimension, xarray_assign, pcAccessor,
    xarray_sel,
    xarray_map,
)

_ANYVAR = r'.*?[^_]+?'  # lazy match (as few characters as possible), anything not ending with '_'
        # could use 'f' strings below to plug in ANYVAR.
        # Instead, copy-pasted the string, to make it easier to search the codebase for these patterns,
        # and to avoid complications with using the '{' and '}' elsewhere in the string.

VECTOR_ZERO = xr.DataArray([0,0,0], coords={'component': XYZ})


''' --------------------- Vector Arithmetic Helper Methods --------------------- '''
# note - these are also provided as staticmethods in VectorArithmeticLoader.
#    thus, e.g., can use PlasmaCalculator.dot_product, if you prefer.
# also note - PlasmaCalculator methods using these functions should prefer the syntax self.dot_product,
#    instead of just using dot_product directly, so that subclasses can overwrite these methods if desired.

@pcAccessor.register('dot_product', aliases=['dot'])
def dot_product(A, B):
    '''return dot product of vectors A and B, assuming vector components along the dimension 'component'.'''
    # note: need to ensure to only sum across matching components
    #    (can't just take_along_dimension('component') without checking if components match)
    # also note: (A * B).sum('component') fails when component is a non-dimension coordinate,
    #    e.g. when there is only 1 component.
    A_dim = 'component' in A.dims
    B_dim = 'component' in B.dims
    # case: both have 'component' dimension. Easy - let xarray handle all the bookkeeping.
    if A_dim and B_dim:
        result = A * B
        return result.sum('component')
    # case: exactly one has 'component' dimension but the other does not.
    elif A_dim or B_dim:
        if A_dim:
            a, b = A, B
        else:
            a, b = B, A
        # now, 'a' has 'component' dimension while 'b' does not.
        if b.coords['component'].size != 1:
            raise NotImplementedError('dot_product, when component is a non-dimension coordinate with size > 1')
        idx = np.nonzero(a.coords['component'].values == b.coords['component'].values)[0]
        if len(idx) == 0:  # no matching components. Result is 0.
            return xr.zeros_like(b).drop_vars('component')  # drop_vars --> result has no 'component' coord.
        else:
            assert len(idx) == 1
            i = idx[0]
            result = a.isel(component=i) * b
            return result.drop_vars('component')  # drop_vars --> result has no 'component' coord.
    # case: A and B both do not have 'component' dimension.
    else:
        # first ensure A and B have the same 'component' coordinate, then it's easy, just multiply them.
        if not (A.coords['component'].size == 1 == B.coords['component'].size):
            raise NotImplementedError('dot_product, when component is a non-dimension coordinate with size > 1')
        A_comps = A.coords['component'].values
        B_comps = B.coords['component'].values
        if np.all(A_comps == B_comps):
            result = A * B
            return result.drop_vars('component')  # drop_vars --> result has no 'component' coord.
        else:
            return xr.zeros_like(A).drop_vars('component')  # drop_vars --> result has no 'component' coord.
    assert False, "if reached this line, made coding mistake."

@pcAccessor.register('magnitude', aliases=['mag', 'mod'])
def magnitude(A, *, squared=False):
    '''return vector magnitude of A, assuming vector components along the dimension 'component'.
    squared: bool, default False
        if True, return |A|**2 instead of |A|.
        [EFF] to get |A|**2, when |A| is not needed,
            magnitude(A, squared=True) is more efficient than magnitude(A)**2
    '''
    result = dot_product(A, A)
    return result if squared else result**0.5

@pcAccessor.register('unit_vector', aliases=['hat', 'dir', 'direction'])
def unit_vector(A):
    '''return a unit vector in the direction of A. Equivalent: A / |A|'''
    return A / magnitude(A)

@pcAccessor.register('angle_xy', aliases=['ang', 'hat2ang', 'angxy', 'dirxy'])
def angle_xy(A):
    '''return angle between +xhat and A, in the xy plane, in radians.
    A should be a DataArray (or Dataset) with x and y in 'component' dimension.
    A can be any vector (does not need to be a unit vector, but it can be.)
    '''
    x = xarray_sel(A, component='x', drop=True)
    y = xarray_sel(A, component='y', drop=True)
    return np.arctan2(y, x)

@pcAccessor.register('angle_xy_to_hat', aliases=['ang2hat', 'angxy2hat', 'dirxy2hat'])
def angle_xy_to_hat(A):
    '''return unit vector u, given angle [radians] between +xhat and u in the xy plane.
    Equivalent: cos(A) * xhat + sin(A) * yhat.
    '''
    return np.cos(A) * XHAT + np.sin(A) * YHAT

def rmscomps(A):
    '''return root mean squared of components of A.
    E.g., rmscomps(A) --> sqrt((Ax^2 + Ay^2 + Az^2) / 3), if A has 3 components.
    '''
    num_comps = A.coords['component'].size  # number of components
    mag2 = magnitude(A, squared=True)
    return np.sqrt(mag2 / num_comps)

@pcAccessor.register
def cross_component(A, B, x, *, yz=None, missing_ok=False):
    '''return x component of A cross B, given A and B which have values for y and z 'component'.

    x: int, str, or Component
        tells component (of result) to get. if int or str, use XYZ.get(x)
    A, B: xarray.DataArray
        vectors to take cross product of.
        must include 'component' dimension including coordinates y and z.
    yz: None or iterable of two (int, str, or Component) objects
        the other two components; (x,y,z) should form a right-handed coordinate system.
        if not provided, infer from x.
    missing_ok: bool, default False
        whether it is okay for 'component' dimension to be missing y or z components, of A or B.
        if True, treat any missing components as 0.
    '''
    if yz is None: yz = YZ_FROM_X[x]
    x = XYZ.get(x)
    default = 0 if missing_ok else UNSET
    Ay, Az = take_along_dimension('component', A, yz, default=default)
    By, Bz = take_along_dimension('component', B, yz, default=default)
    return xarray_assign(Ay * Bz - Az * By, coords={'component': x})

@pcAccessor.register('cross_product', aliases=['cross'])
def cross_product(A, B, *, components=None):
    '''return cross product of vectors A and B, along dimension 'component'.
    If A or B missing any components, treat them as 0.

    components: None or iterable of component specifiers (int, str, or Component)
        tells which components to get.
        None --> get all components (XYZ)
        e.g., (0, 'z') --> get component 0 and component 'z', i.e. X and Z.
    '''
    if components is None:
        components = XYZ
    result = [cross_component(A, B, x, missing_ok=True) for x in components]
    return join_along_dimension('component', result)

def take_perp_to(B, A):
    '''return the component of A perpendicular to B. Equivalent: A - (A dot Bhat) Bhat.
    Note that B is the first argument.
    '''
    Bhat = B / magnitude(B)
    return A - dot_product(A, Bhat) * Bhat

def take_parallel_to(B, A):
    '''return the component of A parallel to B. Equivalent: (A dot Bhat) Bhat.
    Note that B is the first argument.
    '''
    Bhat = B / magnitude(B)
    return dot_product(A, Bhat) * Bhat

# # rephrase take_perp_to and take_parallel_to for xarray accessor (e.g. arr.pc.perp(B)). # #
@pcAccessor.register
def perp(A, B):
    '''return the component of A perpendicular to B. Equivalent: A - (A dot Bhat) Bhat.'''
    return take_perp_to(B, A)

@pcAccessor.register
def parallel(A, B):
    '''return the component of A parallel to B. Equivalent: (A dot Bhat) Bhat.'''
    return take_parallel_to(B, A)

@pcAccessor.register('perpmod', aliases=['perpmag'])
def perpmod(A, B):
    '''return the magnitude of the component of A perpendicular to B. Equivalent: |A - (A dot Bhat) Bhat|'''
    return magnitude(perp(A, B))

@pcAccessor.register('parmod', aliases=['parmag'])
def parmod(A, B):
    '''return the magnitude of the component of A parallel to B. Equivalent: |(A dot Bhat) Bhat|'''
    return magnitude(dot_product(A, unit_vector(B)))


''' --------------------- Vector Arithmetic Loader --------------------- '''

class VectorArithmeticLoader(QuantityLoader):
    '''arithmetic for vectors or components.
    E.g. var_x --> x component of var, i.e. self(var)[..., 0]

    for vector derivatives, see VectorDerivativeLoader.
    '''
    # {var}_{x}; x from characters in 'xyz'; 1 <= len(x) <= 3
    @known_pattern(r'(.*?[^_]+?)_([xyz]{1,3})', deps=[0], ignores_dims=['component'])
    def get_xyz(self, var, *, _match=None):
        '''x, y, and/or z components of var.'''
        base, components = _match.groups()
        cc = [x for x in components] if len(components)>1 else components
        with self.using(component=cc):
            return self(base)

    # # # DOT & CROSS (& magnitude) # # #
    # {var0}_dot_{var1}; optional "__{axes}"
    @known_pattern(r'(.+)_dot_(.+?)(__[xyz]{1,3})?', deps=[0,1], ignores_dims=['component'])
    def get_dot(self, var, *, _match=None, _val0=None, _val1=None, **_known_vals):
        '''dot product. {A}_dot_{B} --> A dot B.
        if component(s) is provided, only include that component(s) during the calculation.
            e.g. A_dot_B__xy --> Ax * Bx + Ay * By.

        [EFF] can provide known vals for A or B, to avoid recalculating them. (include leading underscores.)
            e.g. self('u_dot_E', _u=u, _E=E) --> u dot E, using u and E which are already known.
            if providing value as None, it will be treated as if value not provided.
            CAUTION: Not tested when simultaneously providing components such as A_dot_B__xy.
            can alternatively provide _val0 for A and/or _val1 for B.
        ''' 
        A, B, components = _match.groups()
        Aval = self._provided_val(A, _val0, _known_vals)
        Bval = self._provided_val(B, _val1, _known_vals)
        components = 'xyz' if (components is None) else components[len('__'):]  # remove '__'. e.g. '__x' --> 'x'
        with self.using(component=tuple(components)):
            if Aval is None:
                Aval = self(A)
            if Bval is None:
                Bval = self(B) if A != B else Aval  # [EFF] if A == B, only calculate once.
        return self.dot_product(Aval, Bval)  # use self.dot_product in case subclass overwrites.

    dot_product = staticmethod(dot_product)
    dot = alias('dot_product')

    # mod_{var} or mag_{var}; optional "__{axes}"
    @known_pattern(r'(mod|mag)_(.+?)(__[xyz]{1,3})?', deps=[1], ignores_dims=['component'])
    def get_mod(self, var, *, _match=None, _val0=None, **_known_vals):
        '''magnitude of var. mod_{A} --> |A|.  == sqrt(A dot A) == sqrt(Ax^2 + Ay^2 + Az^2).
        alias: 'mag_{A}' is equivalent to 'mod_{A}'
        if component(s) is provided, only include that component(s) during the calculation.
            e.g. mod_A__xy --> sqrt(Ax^2 + Ay^2).

        [EFF] can provide known vals for {var} to avoid recalculating it. (include leading underscore.)
            e.g. self('mod_E', _E=E) --> |E|, using E which is already known.
            CAUTION: Not tested when simultaneously providing components such as mod_A__xy.
            can alternatively provide _val0 for A.
        '''
        _alias, A, components = _match.groups()   # _alias is 'mod' or 'mag'
        if components is None: components = ''
        vdotv = self(f'{A}_dot_{A}{components}', _val0=_val0, _val1=_val0, **_known_vals)
        return vdotv ** 0.5

    magnitude = staticmethod(magnitude)

    # mod2_{var} or mag2_{var}; optional "__{axes}"
    @known_pattern(r'(mod2|mag2)_(.+?)(__[xyz]{1,3})?', deps=[1], ignores_dims=['component'])
    def get_mod2(self, var, *, _match=None, _val0=None, **_known_vals):
        '''magnitude squared of var. mod2_{A} --> |A|^2.  == A dot A == Ax^2 + Ay^2 + Az^2.
        alias: 'mag2_{A}' is equivalent to 'mod2_{A}'
        if component(s) is provided, only include that component(s) during the calculation.
            e.g. mod2_A__xy --> Ax^2 + Ay^2.

        [EFF] can provide known vals for {var} to avoid recalculating it. (include leading underscore.)
            e.g. self('mod2_E', _E=E) --> |E|**2, using E which is already known.
            CAUTION: Not tested when simultaneously providing components such as mod_A__xy.
            can alternatively provide _val0 for A.
        '''
        _alias, A, components = _match.groups()  # _alias is 'mod2' or 'mag2'
        if components is None: components = ''
        return self(f'{A}_dot_{A}{components}', _val0=_val0, _val1=_val0, **_known_vals)

    # rmscomps_{var}
    @known_pattern(r'rmscomps_(.+)', deps=[0], reduces_dims=['component'])
    def get_rmscomps(self, var, *, _match=None):
        '''root mean squared of components.
        E.g., rmscomps_{A} --> sqrt((Ax^2 + Ay^2 + Az^2) / 3), if A has 3 components.
        '''
        A, = _match.groups()
        Aval = self(A)
        return self.rmscomps(Aval)

    rmscomps = staticmethod(rmscomps)

    # {var0}_cross_{var1}
    @known_pattern(r'(.+)_cross_(.+)', deps=[0,1], ignores_dims=['component'])
    def get_cross(self, var, *, _match=None, _val0=None, _val1=None, **_known_vals):
        '''cross product. {A}_cross_{B} --> A cross B.
        returned components are determined by self.component.
            (see also: the get_xyz pattern. E.g., {A}_cross_{B}_x --> x component of A cross B)

        [EFF] can provide known vals for A or B, to avoid recalculating them. (include leading underscores.)
            e.g. self('u_cross_E', _u=u, _E=E) --> u cross E, using u and E which are already known.
            CAUTION: if providing values, include all self.cross_components_needed() components;
                missing components will be assumed to be 0.
            E.g. if providing u to calculate u_cross_E, but u doesn't have x component, assumes u_x=0.

            can alternatively provide _val0 for A and/or _val1 for B.
        '''
        A, B = _match.groups()
        Aval = self._provided_val(A, _val0, _known_vals)
        Bval = self._provided_val(B, _val1, _known_vals)
        components_needed = self.cross_components_needed()
        with self.using(component=components_needed):
            if Aval is None:
                Aval = self(A)
            if Bval is None:
                Bval = self(B) if A != B else Aval  # [EFF] if A == B, only calculate once.
        components_result = self.component_list()
        return self.cross_product(Aval, Bval, components=components_result)

    cross_component = staticmethod(cross_component)
    
    cross_product = staticmethod(cross_product)
    cross = alias('cross_product')

    def cross_components_needed(self):
        '''return the components vectors need in order to find all cross product components in self.component_list
        e.g. if self.component == 'x', return ('y', 'z'), because (A_cross_B)_x needs Ay, Az, By, Bz but not Ax, Bx.
        '''
        components = self.component_list()  # iterable of Component objects
        if len(components) == 1:
            x = components[0]
            return YZ_FROM_X[x]
        else:
            return tuple(XYZ)
        return YZ_FROM_X[self.component]

    # # # DIRECTIONS # # #
    @known_var(dims=['component'])
    def get_xhat(self):
        '''unit vector in the x direction.
        result components determined by self.component, e.g. xhat_x == 1; xhat_y == 0.
        '''
        c = self.component   # below: sel crashes with tuple but not list.
        return XHAT.sel(component=(list(c) if isinstance(c, tuple) else c))

    @known_var(dims=['component'])
    def get_yhat(self):
        '''unit vector in the y direction.
        result components determined by self.component, e.g. yhat_x == 0; yhat_y == 1.
        '''
        c = self.component   # below: sel crashes with tuple but not list.
        return YHAT.sel(component=(list(c) if isinstance(c, tuple) else c))

    @known_var(dims=['component'])
    def get_zhat(self):
        '''unit vector in the z direction.
        result components determined by self.component, e.g. zhat_x == 0; zhat_z == 1.
        '''
        c = self.component   # below: sel crashes with tuple but not list.
        return ZHAT.sel(component=(list(c) if isinstance(c, tuple) else c))

    # hat_{var}, dir_{var}, direction_{var}, or unit_vector_{var}.
    @known_pattern(r'(hat|dir|direction|unit_vector)_(.+)', deps=[1])
    def get_hat(self, var, *, _match=None, _val0=None, **_known_vals):
        '''unit vector in the direction of var. hat_{A} --> A / |A|.
        returned components are determined by self.component.
            (e.g. when self.component=='x', hat_A --> A_x / |A|.)

        [EFF] can provide known val for A, to avoid recalculating it. (include leading underscore.)
            e.g. self('hat_E', _E=E) --> E / |E|, using E which is already known.
            can alternatively provide _val0 for A.
        '''
        _alias, A = _match.groups()   # _alias is 'hat', 'dir', 'direction', or 'unit_vector'
        Aval = self._provided_val(A, _val0, _known_vals)
        if Aval is None:
            Aval = self(A)
        return Aval / self(f'mod_{A}', _val0=Aval)

    angle_xy = staticmethod(angle_xy)
    angle_xy_to_hat = staticmethod(angle_xy_to_hat)

    # angle_xy_{var}, angxy_{var}, ang_{var}, dirxy_{var}
    @known_pattern(r'(angle_xy|angxy|ang|dirxy)_(.+)', deps=[1], ignores_dims=['component'])
    def get_angle_xy(self, var, *, _match=None, _val0=None, **_known_vals):
        '''angle between +xhat and var, in the xy plane, in radians.

        CAUTION: does not "unwrap"; all angles will be reported in the range -pi to pi.
        See unwrapt2pi for example of unwrapping (see also: np.unwrap)

        angle_xy_{A} --> atan2(Ay, Ax).

        [EFF] can provide known val for A, to avoid recalculating it. (include leading underscore.)
            e.g. self('angle_xy_E', _E=E) --> angle between +xhat and E, using E which is already known.
        '''
        _alias, A = _match.groups()
        Aval = self._provided_val(A, _val0, _known_vals)
        if Aval is None:
            Aval = self(A, component=['x', 'y'])
        return self.angle_xy(Aval)

    @known_pattern(r'unwrapt(2pi)?_(.+)', deps=[1])
    def get_unwrapt_2pi_var(self, var, *, _match=None):
        '''unwrapt_{A} --> unwrapped self(A) along 't', via np.unwrap with period=2*pi.

        CAUTION: result at a given snapshot can vary depending on self.snap,
            (though, (result % 2*pi) will always be the same.)

        E.g. self('unwrapt_angle_xy_E') --> angle between +xhat and E, but unwrapped,
            so e.g. if results change from just above -pi to just below -pi,
            the values below -pi will actually be below -pi,
            instead of being reported as 2*pi + (value just below -pi).
        '''
        _, here = _match.groups()
        val = self(here)
        if 't' not in val.coords:
            raise QuantCalcError(f'cannot unwrapt for val missing "t" coord (coords={list(coords)})')
        tt = val['t']
        if tt.ndim == 0:
            return val  # nothing to unwrap!
        elif tt.ndim == 1:
            tdim = tt.dims[0]
        else:
            raise QuantCalcError(f'cannot unwrapt when val["t"].ndim > 1.')
        # unwrap along the 't' dimension, using np.unwrap with period=2*pi.
        result = xarray_map(val, np.unwrap, axis=tdim, period=2*np.pi)
        return result

    # angle_xy_to_hat_{var}, angxy2hat_{var}, ang2hat_{var}, dirxy2hat_{var}
    @known_pattern(r'(angle_xy_to_hat|angxy2hat|ang2hat|dirxy2hat)_(.+)', deps=[1])
    def get_angle_xy_to_hat(self, var, *, _match=None, _val0=None, **_known_vals):
        '''unit vector u, given angle [radians] between +xhat and u in the xy plane.
        angle_xy_to_hat_{A} --> cos(A) * xhat + sin(A) * yhat.
        '''
        _alias, A = _match.groups()
        Aval = self._provided_val(A, _val0, _known_vals)
        if Aval is None:
            Aval = self(A)
        return self.angle_xy_to_hat(Aval)

    @known_pattern(r'rad2deg_(.+)', deps=[0], ignores_dims=['component'])
    def get_rad2deg(self, var, *, _match=None):
        '''convert radians to degrees. rad2deg_{A} --> A * 180 / pi.
        self('rad2deg_var') == np.rad2deg(self('var'))
        '''
        A, = _match.groups()
        return np.rad2deg(self(A))

    @known_pattern(r'deg2rad_(.+)', deps=[0], ignores_dims=['component'])
    def get_deg2rad(self, var, *, _match=None):
        '''convert degrees to radians. deg2rad_{A} --> A * pi / 180.
        self('deg2rad_var') == np.deg2rad(self('var'))
        '''
        A, = _match.groups()
        return np.deg2rad(self(A))

    # # # ZERO VECTOR # # #
    @known_pattern(r'vector_(\d+)', dims=['component'])  # vector_{n}
    def get_vector_N(self, var, *, _match=None):
        '''vector_n --> vector with n in each component. E.g. vector_0 --> vector with components (0,0,0).
        result components determined by self.component.'''
        n, = _match.groups()
        n = int(n)
        c = self.component
        result = VECTOR_ZERO.sel(component=(list(c) if isinstance(c, tuple) else c))
        return result if n==0 else result + n

    # # # PARALLEL & PERPENDICULAR COMPONENTS # # #
    take_perp_to = staticmethod(take_perp_to)
    take_parallel_to = staticmethod(take_parallel_to)

    # {A}_perp_{B}
    @known_pattern(r'(.+)_perp_(.+)', deps=[0,1])
    def get_perp(self, var, *, _match=None, _val0=None, _val1=None, **_known_vals):
        '''A_perp_B --> A after removing the component of A parallel to B.
        Equivalent to self.take_perp_to(B, A) == A - (A dot Bhat) Bhat.

        [EFF] can provide known vals for A or B, to avoid recalculating them. (include leading underscores.)
            e.g. self('E_perp_B', _E=E, _B=B) --> E perp to B, using E and B which are already known.
            can alternatively provide _val0 for A and/or _val1 for B.
        '''
        A, B = _match.groups()
        Aval = self._provided_val(A, _val0, _known_vals)
        Bval = self._provided_val(B, _val1, _known_vals)
        if Aval is None:
            Aval = self(A, component=None)  # [TODO][EFF] load fewer components here if self.component != XYZ ?
        if Bval is None:
            Bval = self(B, component=None) if A != B else Aval  # [EFF] if A == B, only calculate once.
        result = self.take_perp_to(Bval, Aval)
        if self.component != self.components:  # i.e. only asking for some of the vector components of result
            result = xarray_sel(result, component=self.component)
        return result

    # {A}_par_{B}
    @known_pattern(r'(.+)_par_(.+)', deps=[0,1])
    def get_parallel(self, var, *, _match=None, _val0=None, _val1=None, **_known_vals):
        '''A_par_B --> the component of A parallel to B.
        Equivalent to self.take_parallel_to(B, A) == (A dot Bhat) Bhat.
        see also: A_dot_hat_B, which is equivalent to mod_(A_par_B)

        [EFF] can provide known vals for A or B, to avoid recalculating them. (include leading underscores.)
            e.g. self('E_par_B', _E=E, _B=B) --> E par to B, using E and B which are already known.
            can alternatively provide _val0 for A and/or _val1 for B.
        '''
        A, B = _match.groups()
        Aval = self._provided_val(A, _val0, _known_vals)
        Bval = self._provided_val(B, _val1, _known_vals)
        if Aval is None:
            Aval = self(A, component=None)  # [TODO][EFF] load fewer components here if self.component != XYZ ?
        if Bval is None:
            Bval = self(B, component=None) if A != B else Aval  # [EFF] if A == B, only calculate once.
        result = self.take_parallel_to(Bval, Aval)
        if self.component != self.components:  # i.e. only asking for some of the vector components of result
            result = xarray_sel(result, component=self.component)
        return result

    # {A}_perpmod_{B} or {A}_perpmag_{B}
    @known_pattern(r'(.+)_perp(mod|mag)_(.+)', deps=[0,2], ignores_dims=['component'])
    def get_perpmod(self, var, *, _match=None, _val0=None, _val1=None, **_known_vals):
        '''magnitude of A after removing the component of A parallel to B.
        Equivalent to mod(A_perp_B).

        [EFF] can provide known vals for A or B, to avoid recalculating them. (include leading underscores.)
            e.g. self('E_perpmod_B', _E=E, _B=B) --> |E perp to B|, using E and B which are already known.
            can alternatively provide _val0 for A and/or _val1 for B.
        '''
        A, _mod_or_mag, B = _match.groups()
        with self.using(component=None):  # get A_perp_B across all vector components
            A_perp_B = self(f'{A}_perp_{B}', _val0=_val0, _val1=_val1, **_known_vals)
        return self.magnitude(A_perp_B)

    # {A}_sparmod_{B} or {A}_sparmag_{B}
    @known_pattern(r'(.+)_spar(mod|mag)_(.+)', deps=[0,2], ignores_dims=['component'])
    def get_sparmod(self, var, *, _match=None, _val0=None, _val1=None, **_known_vals):
        '''signed "magnitude" of the component of A parallel to B.
        Equivalent to A_dot_hat_B. Also abs(A_sparmod_B) is equivalent to mod(A_par_B).

        [EFF] can provide known vals for A or B, to avoid recalculating them. (include leading underscores.)
            e.g. self('E_parmod_B', _E=E, _B=B) --> |E par to B|, using E and B which are already known.
            can alternatively provide _val0 for A and/or _val1 for B.
        '''
        A, _mod_or_mag, B = _match.groups()
        if f'_hat_{B}' not in _known_vals:
            _known_vals[f'_hat_{B}'] = self(f'hat_{B}', _val0=_val1, **_known_vals)
        return self(f'{A}_dot_hat_{B}', _val0=_val0, **_known_vals)

    # {A}_parmod_{B} or {A}_parmag_{B}
    @known_pattern(r'(.+)_par(mod|mag)_(.+)', deps=[0,2], ignores_dims=['component'])
    def get_parmod(self, var, *, _match=None, _val0=None, _val1=None, **_known_vals):
        '''magnitude of the component of A parallel to B.
        Equivalent to mod(A_par_B). Also equivalent to abs(A_dot_hat_B)

        [EFF] can provide known vals for A or B, to avoid recalculating them. (include leading underscores.)
            e.g. self('E_parmod_B', _E=E, _B=B) --> |E par to B|, using E and B which are already known.
            can alternatively provide _val0 for A and/or _val1 for B.
        '''
        A, _mod_or_mag, B = _match.groups()
        result = self(f'{A}_sparmod_{B}', _val0=_val0, _val1=_val1, **_known_vals)
        return np.abs(result)



"""
File Purpose: directly loading data from source.
Hookups which read data from files will probably want DirectLoader as a parent.
"""
import os
import re

from .quantity_loader import QuantityLoader

from ..dimensions import SnapHaver, INPUT_SNAP
from ..errors import CacheNotApplicableError
from ..tools import (
    simple_property,
    ProgressUpdater,
    nbytes_path,
)

''' --------------------- DirectLoader --------------------- '''

class DirectLoader(SnapHaver, QuantityLoader):
    '''manages loading data directly from source.

    Not intended for direct use; use a subclass instead.
    '''

    _IMPLIED_IF_NOTES_DIRNAME = '_sim'  # skip including in title if it is the dirname.

    title = simple_property('_title', setdefaultvia='_default_title',
            doc='''title to help distinguish this calculator from others.
            E.g. might want to add self.title to plots.
            Default: os.path.basename(self.dirname) if self.notes_dirname == self.dirname,
                else basename(self.notes_dirname) + basename(self.dirname),
                    except skip basename(self.dirname) if '_sim'.
            Examples of default behavior:
                '/path/to/dir0' --> 'dir0'.
                '/path/to/dir1/_sim' --> 'dir1'.
                '/path/to/dir1/_check0' --> 'dir1_check0'.''')
    def _default_title(self):
        '''default title for self. Here, just returns os.path.basename(self.notes_dirname)'''
        notesdir = self.notes_dirname
        dirname = self.dirname
        if notesdir == dirname:
            return os.path.basename(dirname)
        else:
            if os.path.basename(dirname) == self._IMPLIED_IF_NOTES_DIRNAME:
                return os.path.basename(notesdir) # don't need to show +'_sim' everywhere...
            else:
                return os.path.basename(notesdir) + os.path.basename(dirname)


    # # # FILE PATH INFO # # #
    dirname = simple_property('_dirname',
        doc='''abspath to directory containing data for this DirectLoader.''')
    @dirname.getter
    def dirname(self):
        try:
            return self._dirname
        except AttributeError:
            # make a more helpful error message in case subclass didn't implement dirname yet...:
            raise NotImplementedError(f'{type(self).__name__}.dirname')

    @property
    def nbytes(self):
        '''number of bytes across all files in self.dirname (and subdirectories)'''
        return nbytes_path(self.dirname)

    @property
    def nMbytes(self):
        '''number of megabytes across all files in self.dirname (and subdirectories).
        == self.nbytes / 1024**2
        '''
        return self.nbytes / 1024**2

    @property
    def nGbytes(self):
        '''number of gigabytes across all files in self.dirname (and subdirectories).
        == self.nbytes / 1024**3
        '''
        return self.nbytes / 1024**3


    _INDICATES_NOTES_DIRNAME = {r'_sim.*', r'_check.*'}

    notes_dirname = simple_property('_notes_dirname',
        doc=''''abspath to directory containing plots/notes for a DirectLoader.
        Might be the same directory as self.dirname, but doesn't need to be;
            can explicitly set self.notes_dirname = value, if desired.

        If not set, use notes_dirname == self.dirname,
            unless self.dirname ends with one of the self._INDICATES_NOTES_DIRNAME options.
            E.g. dirname == 'path/to/dir0' --> notes_dirname == 'path/to/dir0'.
            E.g. dirname == 'path/to/dir1/_sim' --> notes_dirname == 'path/to/dir1'.
            E.g. dirname == 'path/to/dir1/_check0' --> notes_dirname == 'path/to/dir1'.

        See also: self.unique_notes_dirname''')
    @notes_dirname.getter
    def notes_dirname(self):
        try:
            return self._notes_dirname
        except AttributeError:
            pass  # get the default.
        dirname = self.dirname
        base = os.path.basename(dirname)
        for pattern in self._INDICATES_NOTES_DIRNAME:
            if re.fullmatch(pattern, base):
                return os.path.dirname(dirname)
        else:  # match not found in for loop
            return dirname

    _INDICATES_UNIQUE_NOTES_DIRNAME = {'_sim'}

    unique_notes_dirname = simple_property('_unique_notes_dirname',
        doc='''abspath to directory containing plots/notes for a DirectLoader,
        probably not shared by DirectLoaders corresponding to any other data.
        (E.g. if

        Might be the same directory as self.dirname, but doesn't need to be;
            can explicitly set self.notes_dirname = value, if desired.

        If not set, use unique_notes_dirname = self.dirname,
            unless dirname ends with one of the self._INDICATES_UNIQUE_NOTES_DIRNAME options.
            E.g. dirname == 'path/to/dir0' --> unique_notes_dirname == 'path/to/dir0'.
            E.g. dirname == 'path/to/dir1/_sim' --> unique_notes_dirname == 'path/to/dir1'.
            E.g. dirname == 'path/to/dir1/_check0' --> unique_notes_dirname == 'path/to/dir1/_check0'.

        See also: self.notes_dirname''')
    @unique_notes_dirname.getter
    def unique_notes_dirname(self):
        try:
            return self._unique_notes_dirname
        except AttributeError:
            pass  # get the default.
        dirname = self.dirname
        base = os.path.basename(dirname)
        for pattern in self._INDICATES_UNIQUE_NOTES_DIRNAME:
            if re.fullmatch(pattern, base):
                return os.path.dirname(dirname)
        else:  # match not found in for loop
            return dirname

    @property
    def snapdir(self):
        '''directory containing the snapshot files.
        [Not implemented by DirectLoader; subclass should implement]
        '''
        raise NotImplementedError(f'{type(self).__name__}.snapdir')

    def snap_filepath(self, snap=None):
        '''convert snap to full file path for this snap.
        snap: None, str, int, or Snap
            the snapshot to load. if None, use self.snap.

        [Not implemented by DirectLoader; subclass should implement]
        '''
        raise NotImplementedError(f'{type(self).__name__}.snap_filepath()')


    # # # DIRECT LOADING-RELATED INFO # # #
    def directly_loadable_vars(self, snap=None):
        '''return tuple of directly loadable variables for this snap.
        snap: None, str, int, or Snap
            the snapshot to load. if None, use self.snap.

        [Not implemented by DirectLoader; subclass should implement]
        '''
        raise NotImplementedError(f'{type(self).__name__}.directly_loadable_vars()')


    # # # LOAD FROMFILE # # #
    # [EFF] whether to slice directly when loading from files, if using self.slices.
    _slice_maindims_in_load_direct = False
    # if True, MainDimensionsHaver logic expects DirectLoader subclass to handle slicing
    #   in its implementation of load_fromfile(). Using self._slice_maindims_numpy might help.

    def _var_for_load_fromfile(self, varname_internal):
        '''return var, suitably adjusted to pass into load_fromfile().
        Any adjustments are acceptable, as long as load_fromfile() understands the result.

        Example: might want to adjust var based on self._loading_{dim}, e.g.:
            if getattr(self, '_loading_component', False):
                return varname_internal + str(self.component)
            # e.g. 'b' --> 'bz', when loading across 'component' dim and self.component=='z'.

        The implementation here just returns varname_internal, unchanged.
        Subclass might want to override.
        '''
        return varname_internal

    def load_fromfile(self, fromfile_var, *args, **kw):
        '''load fromfile_var directly from file, returning raw data.
        E.g. result might be a numpy array with dims corresponding to self.maindims.

        (if called load_direct() via self.load_maindims_var_across_dims,
            then result should definitely have maindims dims.
            But it is also perfectly acceptable to have other dims.)

        (To hook these results into the main PlasmaCalcs architecture,
            define a BasesLoader inheriting from AllBasesLoader & SimpleDerivedLoader.
            See hookups subpackage for examples.)

        [Not implemented by DirectLoader; subclass should implement]
        '''
        raise LoadingNotImplementedError(f'{type(self).__name__}.load_fromfile({fromfile_var!r})')


    # # # LOAD INPUT # # #
    def load_input(self, fromfile_var, *args, **kw):
        '''load value of fromfile_var but when self.snap is INPUT_SNAP.
        [Not implemented for this subclass; loading direct when self.snap is INPUT_SNAP will crash]
        (optional; subclass can override in order to implement INPUT_SNAP functionality.
            see EppicCalculator.load_input() for an example.)
        '''
        errmsg = f'{type(self).__name__}.load_input() not implemented.'
        if self.snap is INPUT_SNAP:
            errmsg += ' (this causes load_direct() to fail when self.snap is INPUT_SNAP)'
        raise LoadingNotImplementedError(errmsg)


    # # # LOAD DIRECT # # #
    # [TODO][MV] consider moving the load_direct() implementation here to QuantityLoader instead.
    def load_direct(self, var, *args, **kw):
        '''load var "directly", either from a file, self.setvars, self.direct_overrides,
        or via self.load_input() if self.snap is INPUT_SNAP.

        Steps:
            1) attempt to get var from cache or self.setvars.
                [EFF] only tries this if we are not self._inside_quantity_loader_call_logic,
                to avoid redundant calls to self.get_set_or_cached.
            2) adjust var name if appropriate, via new_varname = self._var_for_load_fromfile(var).
                if new_varname != old varname, attempt to get new_varname from cache or setvars.
            3) if self.snap is INPUT_SNAP, return self.load_input(new_varname).
            4) super().load_direct(adjusted_var, *args, **kw),
                which will use self.load_fromfile(...) unless any overrides apply here.
        '''
        # (0) - bookkeeping; delete any previous value of self._load_direct_used_override
        try:
            del self._load_direct_used_override
        except AttributeError:
            pass   # that's fine. Just want to reset it before running this function.
        # (1)
        if not getattr(self, '_inside_quantity_loader_call_logic', False):
            try:
                result = self.get_set_or_cached(var)
            except CacheNotApplicableError:
                pass
            else:
                self._load_direct_used_override = var
                return result
        # (2)
        fromfile_var = self._var_for_load_fromfile(var)
        if fromfile_var != var:  # check setvars & cache
            try:
                result = self.get_set_or_cached(fromfile_var)
            except CacheNotApplicableError:
                pass
            else:
                self._load_direct_used_override = fromfile_var
                return result
        # (3)
        if self.snap is INPUT_SNAP:
            return self.load_input(fromfile_var, *args, **kw)
        # (4)
        return super().load_direct(fromfile_var, *args, **kw)


    # # # KNOWN VARS # # #
    @known_pattern(r'load_(.+)', dims=['snap'])  # load_{var} pattern
    def get_load_direct_maindims_var(self, var, *, _match=None):
        '''load maindims var directly (from files, cache, or setvars; see self.load_direct()).
        var should be name of a directly loadable var (see self.directly_loadable_vars()).
        Output always uses 'raw' units, regardless of self.units,
            but coords are in self.coords_units (default: same as self.units).
        '''
        # [TODO] doesn't currently interact well with other patterns, but it should...
        #       (workaround: use parenthesis e.g. load_tg/m fails but (load_tg)/m is fine)
        here, = _match.groups()
        with self.using(coords_units=self.coords_units_explicit, units='raw'):
            result = self.load_maindims_var_across_dims(here, dims=['snap'])
        return result


    # # # CONVENIENT MISC. METHODS (subclass doesn't need to override) # # #
    def _check_files_readable(self):
        '''check that all files in self.dirname are readable (checks recursively inside directories too).
        return number of files checked, number of files readable, and number of files not readable.

        print exceptions along the way if any files are not readable!
        '''
        total = len([f for _, _, files in os.walk(self.dirname) for f in files])
        successes = 0
        failures = 0
        i = 0
        updater = ProgressUpdater(1, wait=True)  # at most 1 update per second; wait before first printout.
        for dirpath, dirnames, filenames in os.walk(self.dirname):
            for filename in filenames:
                i += 1
                updater.print(f"Checking files... at {i} out of {total}")
                filepath = os.path.join(dirpath, filename)
                try:
                    with open(filepath, 'rb') as f:  # read 1 bit of file to check if it is readable.
                        f.read(1)
                except Exception as e:
                    print(f'Error reading {filepath}: {e}')
                    failures += 1
                else:
                    successes += 1
        updater.finalize(f"Checking files")
        return successes + failures, successes, failures



"""
File Purpose: base quantities for EppicCalculator
"""
import re

import numpy as np
import xarray as xr

from .eppic_input_deck import EppicNeutral, EppicNeutralList
from ...dimensions import INPUT_SNAP
from ...errors import (
    InputConflictError, FluidValueError, DimensionAttributeError,
    FormulaMissingError, LoadingNotImplementedError,
)
from ...quantities import AllBasesLoader, SimpleDerivedLoader
from ...tools import UNSET, is_iterable_dim, simple_property


class EppicBasesLoader(AllBasesLoader, SimpleDerivedLoader):
    '''base quantities based on Eppic output.'''

    # # # BASE_QUANTS directly from eppic # # #

    @known_var(load_across_dims=['fluid'])
    def get_m(self):
        '''mass, of a "single particle". For protons, ~= +1 atomic mass unit'''
        return xr.DataArray(self.fluid.m * self.u('M'), attrs=self.units_meta())

    @known_var(load_across_dims=['fluid'])
    def get_q(self):
        '''charge, of a "single particle". For protons, == +1 elementary charge'''
        return xr.DataArray(self.fluid.q * self.u('q'), attrs=self.units_meta())

    @known_var(load_across_dims=['fluid'])
    def get_nusn(self):
        '''collision frequency (of self.fluid) with neutrals.
        "frequency for one particle of species s (self.fluid) to collide with any of the neutrals."
        [TODO] how does this compare with simulated collisions in PIC?
        '''
        return xr.DataArray(self.fluid['coll_rate'] * self.u('Hz'), attrs=self.units_meta())

    @known_var(load_across_dims=['component'], aliases=['dspace'])
    def get_ds(self):
        '''grid spacing (of output files). vector(ds), e.g. [dx, dy, dz]. Depends on self.component.
        ds = (ds_sim / nout_avg) * subsampling_step. subsampling_step=1 unless subsampling_info exists.
        '''
        x = str(self.component)
        dx_raw = self.input_deck.get_dspace(x) * self._get_subsampling_step(x)
        dx = dx_raw * self.u('length')
        return xr.DataArray(dx, attrs=self.units_meta())

    @known_var(deps=['ds_sim'], ignores_dims=['component'])
    def get_ds_for_timescales(self):
        '''ds used when calculating timescales. vector(ds), e.g. [dx, dy, dz].
        Like ds_sim, the value here doesn't include the nout_avg factor.

        ds_for_timescales equals ds_sim for components corresponding to self.input_deck.maindims()
            (regarless of current value of self.component).
        '''
        return self('ds_sim', component=self.input_deck.maindims())

    # # # loading directly from eppic # # #

    @known_var(dims=['snap', 'fluid'], aliases=['den'])
    def get_deltafrac_n(self):
        '''normalized density perturbation. (directly from Eppic.) deltafrac_n = (n - n0) / n0.'''
        return self.load_maindims_var_across_dims('den', dims=['snap', 'fluid'])  # [dimensionless]

    @known_var(load_across_dims=['fluid'])
    def get_n0(self):
        '''background density. (directly from Eppic.)
        note: as with all other quantities, this will be output in [self.units] unit system;
            numerically equal to eppic.i value if using 'raw' units.
        '''
        try:
            n0 = self.fluid.n0
        except AttributeError:
            errmsg = f'{type(self.fluid).__name__} object has no attribute "n0".'
            raise DimensionAttributeError(errmsg) from None
        return xr.DataArray(n0 * self.u('n'), attrs=self.units_meta())

    @known_var(dims=['snap', 'fluid', 'component'])
    def get_flux(self):
        '''flux. (directly from Eppic)'''
        return self.load_maindims_var_across_dims('flux', u='flux', dims=['snap', 'fluid', 'component'])

    @known_var(dims=['snap', 'fluid', 'component'])
    def get_nvsqr(self):
        '''n * ({x} component of velocity)^2. (directly from Eppic)'''
        return self.load_maindims_var_across_dims('nvsqr', u='n u2', dims=['snap', 'fluid', 'component'])

    @known_var(dims=['snap'])
    def get_phi(self):
        '''electric potential. (directly from Eppic)'''
        return self.load_maindims_var_across_dims('phi', u='E length', dims=['snap'])

    @known_var(load_across_dims=['component'])
    def get_E_ext(self):
        '''external electric field. (directly from Eppic).'''
        x = str(self.component)
        Ex = self.input_deck.get(f'E{x}0_external', 0) * self.u('E')
        return xr.DataArray(Ex, attrs=self.units_meta())

    @known_var(dims=['snap', 'fluid'])
    def get_vdist(self):
        '''velocity distribution. (directly from Eppic)'''
        if self.current_n_fluid() != 1:
            raise FluidValueError("get_vdist doesn't work with multiple fluids. Maybe you forgot to choose your fluid?")
        return self.load_across_dims(self._vdist_loader, dims=['snap', 'fluid'])

    def _vdist_loader(self):
        '''load vdist. Helper function for self.get_vdist.'''
        result = self.load_direct("vdist")
        if self._load_direct_used_override is not None:  # got var from an override, not the file.
            return result
        # else, got result directly from a file.
        result_length = len(result)
        target_length = self.fluid["pnvx"]
        if result_length > target_length: # FOR OLD VERSION OF EPPIC WITH PHDF OUTPUT_VDIST BUG
            result = np.sum(np.split(result, result_length//target_length), axis=0)
        result = self.assign_velocity_coords(result)
        return self.record_units(result)

    @known_var(load_across_dims=['component'])
    def get_B(self):
        '''magnetic field. (directly from Eppic)'''
        x = str(self.component)
        Bx = self.input_deck.get(f'B{x}', 0) * self.u('b')
        return xr.DataArray(Bx, attrs=self.units_meta())


    # # # INPUT_SNAP CASE # # #
    # (load_direct() redirects to here when self.snap is INPUT_SNAP)
    def load_input(self, fromfile_var, *args__None, **kw__None):
        '''load input value -- one of the base quantities (fromfile_var) directly from eppic.i file.
        Results are always in 'raw' units.
        E.g., self('den', fluid=0) internally uses self.load_input('den0') for INPUT_SNAP,
            and self.load_fromfile('den0') for all other snaps.

        Possible fromfile_vars are:
            'phi': electric potential, excluding any imposed E_ext.
                Always 0. (== "perturbation" to E_ext, implied by eppic.i file)
            'den{N}': density of dist N.
                Always 0. (== "perturbation" to n, implied by eppic.i file)
            'flux{x}{N}': flux of dist N in x direction.
                fN.get_n0() * fN.get_v0(x), where fN = self.fluids.get(N).
            'nvsqr{x}{N}': n * (v_x^2) of dist N in x direction.
                fN.get_nvsqr0(x), where fN = self.fluids.get(N).
                == fN.get_n0() * (fN.get_v0(x)**2 + fN.get_vth0(x)**2)
        '''
        pattern = r'(phi|den(\d+)|flux([xyz])(\d+)|nvsqr([xyz])(\d+))'
        match = re.fullmatch(pattern, fromfile_var)
        if match is None:
            errmsg = (f'fromfile_var={fromfile_var!r} not recognized. '
                      'Expected var like: phi, den{N}, flux{x}{N}, or nvsqr{x}{N}.')
            raise LoadingNotImplementedError(errmsg)
        var = match.groups()[0]
        if var == 'phi':
            result = 0.0
        elif var.startswith('den'):
            result = 0.0
        elif var.startswith('flux'):
            x, N = re.fullmatch(r'flux([xyz])(\d+)', var).groups()
            fN = self.fluids.get(int(N))
            result = fN.get_n0() * fN.get_v0(x)
        elif var.startswith('nvsqr'):
            x, N = re.fullmatch(r'nvsqr([xyz])(\d+)', var).groups()
            fN = self.fluids.get(int(N))
            result = fN.get_nvsqr0(x)
        else:
            assert False, 'coding error if reached this line'
        return xr.DataArray(result)  # in [raw] units.

    @known_pattern(r'frominputs_(.+)', deps=[0])
    def get_frominputs_value(self, var, *, _match=None):
        '''compute var based on inputs, i.e. when self.snap = INPUT_SNAP.
        result drops 'snap' coord, if it exists.

        self('frominputs_var') is equivalent to:
        with self.using(snap=INPUT_SNAP):
            return self('var').drop_vars('snap', errors='ignore')
        '''
        here, = _match.groups()
        with self.using(snap=INPUT_SNAP):
            result = self(here)
        return result.drop_vars('snap', errors='ignore')


    # # # QUASINEUTRAL CASE # # #
    # (this belongs in eppic hookup, not quasineutral.py, because it depends on eppic.)

    def _qn_direct_den0(self):
        '''gets den0 in quasineutral case.'''
        self.assert_QN()
        return self('deltafrac_ne')
        # note - 'get_deltafrac' rule is in StatsLoader; 'get_ne' rule is in QuasineutralLoader.

    def direct_overrides_dynamic(self):
        '''returns dict of {var: override} for all overrides of self which depend on behavior_attrs of self.'''
        result = super().direct_overrides_dynamic()
        if self.quasineutral:
            result.update({'den0': type(self)._qn_direct_den0})
        return result
    

    # # # HELPER QUANTS derived from eppic values # # #

    @known_var(deps=['n', 'nvsqr'])
    def get_vsqr(self):
        '''v^2 in each grid cell. components given by: v_x^2 = nvsqr_x / n.'''
        return self('nvsqr') / self('n')

    @known_pattern(r'Ta(joule)?(_global)?',
                   deps=['m', 'vsqr', {1: lambda groups: 'nmean_u' if groups[1] else 'u'}])
    def get_Ta_or_Tajoule(self, var, *, _match=None):
        '''temperature ("anisotropic").
        'Ta'      --> (m * (vsqr - u^2) / kB) [Kelvin]
        'Tajoule' --> (m * (vsqr - u^2))      [energy units]
        'Ta_global' or 'Tajoule_global'
            --> use nmean_u instead of u. nmean_u is the density-weighted mean velocity,
                i.e. it is the mean velocity across all particles in the box
                (which is not necessarily equivalent to mean across all cells in the box).
                Note that if reading u directly from simulation, nmean_u should be equal to moment1.
        components given by:
            Ta_x      = m * (vsqr_x - u_x^2) / kB
            Tajoule_x = m * (vsqr_x - u_x^2)

        There are 3 different quantities to consider, related to averaging:
            - Ta_from_moment2_x: single value, temperature if the entire box is one distribution;
                    this is probably what observations would see if they can't resolve the box.
                    Equal to x component of m * <(v - <v>)^2>, where <*> indicates mean across all particles.
                    This equals m * (<v^2> - 2<v*<v>> + <v>^2) = m * (<v^2> - <v>^2).
                    Note: <v^2> == nmean_vsqr, because nmean is equivalent to "mean across particles".
            - Ta_global_x: one value per grid cell. Not defined as a "temperature",
                    instead, defined as the quantity whose nmean is equivalent to Ta_from_moment2.
                    Ta_global_x = m * (vsqr - <v>^2) == m * (vsqr - nmean_u^2).
            - Ta_x: one value per grid cell. Temperature for that grid cell;
                    this is probably what observations would see if they could resolve the grid cells.
                    Ta_x = m * (vsqr - u^2), where u is the average of particle velocities in that cell, only.
                    Because Ta_x doesn't use global nmean_u, nmean_Ta_x is NOT equivalent to Ta_from_moment2_x.
        '''
        here, global_ = _match.groups()
        # note: nmean_u should equal moment1, but might be different if used self.set_var to set u or n;
        # that is why 'nmean_u' is used here instead of 'moment1'.
        u = self('nmean_u') if global_ else self('u')
        result = self('m') * (self('vsqr') - u**2)
        if here != 'joule':  # Ta, not Tajoule
            result = result / self.u('kB')
        return result

    @known_var(deps=['grad_phi'])
    def get_E_phi(self):
        '''electric field, from phi. E_phi = -grad(phi). Doesn't include E_ext component.'''
        return -self('grad_phi')


    # # # BASE_QUANTS derived from eppic values # # #

    @known_var(deps=['n0', 'deltafrac_n'])
    def get_n(self):
        '''number density. n = n0 * (1 + deltafrac_n)'''
        # [TODO][EFF] implement ne separately, if quasineutral.
        #   presently, implementation will work but get_n gets deltafrac_ne from ne,
        #       (since deltafrac_ne is treated as the "directly-loaded" var)
        #       then gets ne from deltafrac_ne.
        #   Fixing this is probably a very low priority,
        #       mostly because doing it cleanly involves altering some of the dimension tools,
        #       but here's a note about it in case we have time.
        if self._fluid_is_neutral():
            return self._get_n_neutral()
        n0 = self('n0')
        dfn = self('deltafrac_n')
        return n0 * (1 + dfn)

    @known_var(deps=['flux', 'n'])
    def get_u(self):
        '''velocity. u = flux / n'''
        if self._fluid_is_neutral():
            return self('u_neutral')  # always 0 in every component, for EPPIC.
        return self('flux') / self('n')

    cls_behavior_attrs.register('T_indim_only', default=False)
    T_indim_only = simple_property('_T_indim_only', default=False,
        doc='''whether to consider only components in maindims when getting T from Ta.
        Irrelevant in 3D; 3D always has self('T') == sqrt((1/3)*(Ta_x^2 + Ta_y^2 + Ta_z^2)).
        In 2D x-y sim, True --> self('T') == self('T_indim') == sqrt((1/2)*(Ta_x^2 + Ta_y^2)).
            2D & False --> self('T') = same formula as in 3D sim.''')

    @known_var(attr_deps=[('T_indim_only', {True: 'T_indim', False: 'rmscomps_Ta'})])
    def get_T(self):
        '''temperature, from nv^2. Rms average of all (3) anisotropic temperature components.
        T == sqrt((Ta_x^2 + Ta_y^2 + Ta_z^2)/3)

        if self.T_indim_only, instead return self('T_indim'), which ignores Tz if data is 2D.
        '''
        if self.T_indim_only:
            return self('T_indim')
        if self._fluid_is_neutral():
            return self._get_T_neutral()
        return self('rmscomps_Ta', component=None)

    @known_var(deps=['rmscomps_Ta'])
    def get_T_indim(self):
        '''temperature, from nv^2. Rms average of Ta components in self.maindims.
        Equivalent to self('T') in 3D, but in 2D xy data, T_indim ignores Tz.
        '''
        if self._fluid_is_neutral():
            return self._get_T_neutral()
        return self('rmscomps_Ta', component=self.maindims)

    @known_var(deps=['E_ext', 'E_phi'])
    def get_E(self):
        '''electric field. E = E_external + E_phi = E_external - grad(phi)'''
        E_ext = self('E_ext')
        E_phi = self('E_phi')
        return E_ext + E_phi


    # # # HANDLING NEUTRALS # # #

    def _fluid_is_neutral(self):
        '''returns True if self.fluid is neutral (either one neutral or multiple).
        True if self.fluid is an EppicNeutral, or iterable with only EppicNeutrals.
        '''
        f = self.fluid
        if isinstance(f, (EppicNeutral, EppicNeutralList)):
            return True
        # included EppicNeutralList check above, for efficiency.
        # (code below would also work for EppicNeutralList, but it's slower)
        # now, need to check if fluid is iterable and all are neutrals
        #  (e.g. if fluid is a list (but not FluidList) of only neutrals)
        if self.fluid_is_iterable() and all(isinstance(f_, EppicNeutral) for f_ in f):
            return True
        return False

    @known_var(aliases=['u_n'])
    def get_u_neutral(self):
        '''returns 0, since u_neutral is always 0 in every component, for eppic.'''
        return self('vector_0')

    @known_var(load_across_dims=['fluid'], aliases=['m_n_f'])
    def get_m_neutral_f(self):
        '''mass, of a "single neutral particle," varying across fluid. For Hydrogen, ~= +1 atomic mass unit'''
        if "massd_neutral" in self.fluid.params:
            return xr.DataArray(self.fluid['massd_neutral'] * self.u('M'), attrs=self.units_meta())
        else:
            return self.get_neutral('m')
    
    @known_var(aliases=['m_n'])
    def get_m_neutral(self):
        '''mass, of a "single neutral particle". For Hydrogen, ~= +1 atomic mass unit'''
        for fluid in self.fluid_list():
            if "massd_neutral" in fluid.params:
                return self.get_m_neutral_f()
        return self.get_neutral('m')

    @known_var(aliases=['n_n'])
    def get_n_neutral(self):
        '''number density of neutrals.'''
        # [TODO][REF] reduce redundancy with _get_n_neutral below?
        return self.get_neutral('n')

    @known_var(aliases=['T_n'])
    def get_T_neutral(self):
        '''temperature of neutrals.'''
        # [TODO][REF] reduce redundancy with _get_T_neutral below?
        return self.get_neutral('T')

    # not a known_var; just a helper function. called by self.get_n if appropriate.
    def _get_n_neutral(self):
        '''get the number density of self.fluid when it is an EppicNeutral or EppicNeutralList.
        (Or, crash with a helpful FormulaMissingError error message.)
        '''
        if not self._fluid_is_neutral():
            errmsg = (f'To use _get_n_neutral, self.fluid must neutral; but got fluid={self.fluid}')
            raise TypeError(errmsg)
        if getattr(self, '_inside_logic_of_get_n_neutral', False):
            with self.using(_inside_logic_of_get_n_neutral=True):
                # ^ prevent recursion on this part, in case self('n') calls self._get_n_neutral()
                try:
                    return self('n')
                    # if successful, user has set n using set_var. Probably via set_var('nj', value, ...)
                except DimensionAttributeError:  # crashed in get_n0; fluid has no 'n0' attribute.
                    pass  # handled below.
        errmsg = (f"Cannot get 'n' for neutral self.fluid={self.fluid}, when no value was set.\n"
                  f"Maybe you forgot to set_var('nj', value, ...) or set_var('n', value, ...)?")
        raise FormulaMissingError(errmsg) from None

    # not a known_var; just a helper function. called by self.get_T if appropriate.
    def _get_T_neutral(self):
        '''get the temperature of self.fluid when it is an EppicNeutral or EppicNeutralList.'''
        if not self._fluid_is_neutral():
            errmsg = (f'To use _get_T_neutral, self.fluid must neutral; but got fluid={self.fluid}')
            raise TypeError(errmsg)
        #with self.using(ncpu=1):  # never need multiprocessing for Tn_loader.
        return self.load_across_dims(self._T_neutral_loader, dims=['fluid'])

    def _T_neutral_loader(self):
        '''load T for neutrals. Helper function for self._get_T_neutral.
        gets T from vth and m. vth = sqrt(kB * T / m) --> T = m * vth^2 / (kB)
        '''
        vth = self.fluid.vth * self.u('u')
        if vth is None:
            errmsg = f'Cannot get Tn when {type(self.fluid).__name__}.vth not provided (got vth=None)'
            raise FluidValueError(errmsg)
        m = self.fluid.m * self.u('mass')
        T = m * vth**2 / self.u('kB')
        return xr.DataArray(T, attrs=self.units_meta())


    # # # --- SETTING VALUES; KNOWN SETTERS --- # # #
    # used when using set_var.

    @known_setter
    def set_n(self, value, **kw):
        '''set n to this value, by setting 'den' to the appropriate value.
        (To set neutral (jfluid) density, use set_nj instead.)
        '''
        n0 = self('n0')
        den = (value - n0) / n0
        return self.set('den', den, **kw)

    @known_setter
    def set_den(self, value, **kw):
        '''set den to this value. See also: set_n'''
        return self.set_var_internal('den', value, ['snap', 'fluid'], **kw, ukey='1')  # [dimensionless] --> ukey='1'

    @known_setter
    def set_u(self, value, **kw):
        '''set u to this value, by setting 'flux' to the appropriate value. flux = n * u.
        Depends on the current value of n; if also setting n be sure to set n first.
        '''
        n = self('n')
        flux = n * value
        return self.set('flux', flux, **kw)

    @known_setter
    def set_flux(self, value, **kw):
        '''set flux to this value. See also: set_u'''
        return self.set_var_internal('flux', value, ['snap', 'fluid', 'component'], **kw, ukey='flux')

    @known_setter
    def set_T(self, value, **kw):
        '''set T to this value, by setting all components of 'nvsqr' to the appropriate value.
        kB T ==  m * [vsqr - u^2]. --> nvsqr = n * ((kB * T / m) + u^2)
        Depends on the current value of n; if also setting n be sure to set n first.
        (To set neutral (jfluid) temperature, use set_Tj instead.)
        '''
        n = self('n')
        m = self('m')
        vtherm2 = self.u('kB') * value / m
        result = []
        for _component in self.iter_components():
            nvsqr = n * (vtherm2 + self('u')**2)
            result.extend(self.set('nvsqr', nvsqr, **kw))
        return result

    @known_setter
    def set_nvsqr(self, value, **kw):
        '''set nvsqr to this value. See also: set_T'''
        return self.set_var_internal('nvsqr', value, ['snap', 'fluid', 'component'], **kw, ukey='n u2')

    @known_setter
    def set_phi(self, value, **kw):
        '''set phi to this value.'''
        return self.set_var_internal('phi', value, ['snap'], **kw, ukey='E length')

    # NOT A KNOWN SETTER - combined method for setting multiple things.
    def set_bases(self, *, n=UNSET, u=UNSET, T=UNSET, ux=UNSET, uy=UNSET, uz=UNSET, forall=[],
                  v=UNSET, vx=UNSET, vy=UNSET, vz=UNSET, **kw):
        '''set n, u, and T to these values (for the relevant current behavior in self..).
        (e.g. if self.fluid=0, setting values for fluid=0, only.)
        
        n, u, T, ux, uy, uz: UNSET, None, or number
            UNSET --> don't set this value
            None --> delete this value from self.setvars
            number --> set this value.
            Note: if u is provided, self.component must be a single value.
        v, vx, vy, vz:
            aliases for u, ux, uy, uz, respectively.
        forall: list of strings
            behavior attrs to which set value should apply to all of.
            E.g., forall=['snap'] --> set value applies at all snaps instead of just current snap.

        additional kwargs, if provided, go to self.using(**kw) during the operation.

        returns (list of set quantities, list of unset quantities)
        '''
        # aliases / bookkeeping
        for v_, u_ in zip([v, vx, vy, vz], [u, ux, uy, uz]):
            if v_ is not UNSET and u_ is not UNSET:
                raise InputConflictError(f'set v or u or neither, but not both.')
        if v is not UNSET: u = v
        if vx is not UNSET: ux = vx
        if vy is not UNSET: uy = vy
        if vz is not UNSET: uz = vz
        # actually setting values
        added = []
        deled = []
        with self.using(**kw):
            # set n first; setting u and T internally depends on n.
            if n is UNSET:
                pass
            elif n is None:
                deled.extend(self.unset('den'))
            else:
                added.extend(self.set('n', n, forall=forall))
            # set u
            if u is UNSET:
                pass
            elif u is None:
                deled.extend(self.unset('flux'))
            else:
                added.extend(self.set('u', u, forall=forall))
            # set u, component-wise
            for x, ux_ in zip('xyz', [ux, uy, uz]):
                with self.using(component=x):
                    if ux_ is UNSET:
                        pass
                    elif ux_ is None:
                        deled.extend(self.unset(f'flux'))
                    else:
                        added.extend(self.set(f'u', ux_, forall=forall))
            # set T
            if T is UNSET:
                pass
            elif T is None:
                deled.extend(self.unset('nvsqr'))
            else:
                added.extend(self.set('T', T, forall=forall))
        return (added, deled)

    # # # SETTERS FOR JFLUID # # #
    @known_setter
    def set_nj(self, value, **kw):
        '''set n to this value for the current self.jfluid.
        Getting 'n' returns value when self.fluid is the self.jfluid from when 'n' was set.
        The implementation here assumes jfluids are inherently different from fluids,
            which is true for Eppic, where jfluid is neutral (by default).
        '''
        if is_iterable_dim(self.jfluids) and is_iterable_dim(self.fluids):
            if any(jfluid in self.fluids for jfluid in self.jfluids):
                raise FluidValueError(f'jfluids and fluids overlap; cannot use set_nj(...).')
        # temporarily set fluid to jfluid. include fluids=self.jfluids because fluids and jfluids differ.
        with self.using(fluids=self.jfluids, fluid=self.jfluid):
            battrs = ['snap', 'fluid']  # 'fluid' because getting 'n' with getj will set fluid=jfluid too.
            return self.set_var_internal('n', value, battrs, **kw, ukey='n')

    @known_setter
    def set_Tj(self, value, **kw):
        '''set T to this value for the current self.jfluid.
        Getting 'n' returns value when self.fluid is the self.jfluid from when 'n' was set.
        The implementation here assumes jfluids are inherently different from fluids,
            which is true for Eppic, where jfluid is neutral (by default).
        '''
        if is_iterable_dim(self.jfluids) and is_iterable_dim(self.fluids):
            if any(jfluid in self.fluids for jfluid in self.jfluids):
                raise FluidValueError(f'jfluids and fluids overlap; cannot use set_Tj(...).')
        # temporarily set fluid to jfluid. include fluids=self.jfluids because fluids and jfluids differ.
        with self.using(fluids=self.jfluids, fluid=self.jfluid):
            battrs = ['snap', 'fluid']
            return self.set_var_internal('T', value, battrs, **kw, ukey='1')  # [K] is "dimensionless"; u_K==1.



"""
File Purpose: handling parenthesis in var names.
"""

from ..quantity_loader import QuantityLoader
from ...errors import QuantCalcError, InputError
from ...tools import BijectiveMemory, UNSET


class ParenthesisLoader(QuantityLoader):
    '''parenthesis.
    parenthesis_memory: BijectiveMemory (subclass of dict).
        {key: var} for all vars ever inside parentheses, across all instances of ParenthesisLoader.
        keys <--> var mapping is one-to-one during each Python session.
        parenthesis_memory.inverse provides the {var: key} mapping.
    '''
    # # # HELPERS # # #
    parenthesis_memory = BijectiveMemory()

    @classmethod
    def _parenthesis_converted_var(cls, before, here, after):
        '''returns the var after converting it from var like '{before}({here}){after}'.
        use same before & after, but use {i} instead of here.
        Looks like: f'{before}{{{key}}}{after}' where key=int, unique to this var.
        '''
        key = cls.parenthesis_memory.key(here)
        return before + '{' + str(key) + '}' + after  # '+' instead of f-string for code-readability with braces.

    def _braced_int_dep(self, _var, groups):
        '''returns the 'dep' associated with this var for _get_braced_int_from_parenthesis_memory'''
        i, = groups
        return self.parenthesis_memory[int(i)]  # if this key doesn't exist, made a coding error.

    def _parenthesis_dep(self, _var, groups):
        '''returns the 'dep' associated with this var for get_parenthesis.'''
        before, here, after = groups
        return self._parenthesis_converted_var(before, here, after)

    # # # GETTABLE PATTERNS # # #
    # '{before}({here}){after}', with no parenthesis inside {here}
    @known_pattern(r'(.*)[(]([^()]+)[)](.*)', deps=[_parenthesis_dep])
    def get_parenthesis(self, var, *, _match=None, **kw):
        '''parenthesis. '{prefix}({here}){suffix}', with no parenthesis inside {here}.
        Ensures that {here} is interpreted as a single expression. Useful when combining operations.

        E.g. 'sqrt_q/m', with no parenthesis, has an ambiguous interpretation.
            'sqrt_(q/m)' unambiguously refers to "sqrt(ratio between q and m)".
            '(sqrt_q)/m' unambiguously refers to "ratio between sqrt(q) and m".
            The no-parenthesis interpretation is subject to change without warning, in future code updates.
            (You can always check the interpretation that is being used, via self.match_var_tree(var))
        '''
        before, here, after = _match.groups()
        newvar = self._parenthesis_converted_var(before, here, after)  # e.g. mean_(u/n) --> mean_{0}
        return self(newvar, **kw)   # the braces will be evaluated later, during self.get_braced_int_from_parenthesis_memory.

    # '{i}' where i is an int.
    @known_pattern(r'[{](\d+)[}]', deps=[_braced_int_dep])
    def get_braced_int_from_parenthesis_memory(self, var, *, _match=None):
        '''{i} --> self(mem_var), where mem_var = self.parenthesis_memory_key_to_var[int(i)].'''
        i, = _match.groups()
        mem_var = self.parenthesis_memory[int(i)]
        return self(mem_var)

    # # # PICKLING # # #
    def __getstate__(self):
        '''return state for pickling. pickle default ignores class attr (parenthesis_memory) but we need it.'''
        if hasattr(super(), '__getstate__'):
            state = super().__getstate__()
        else:
            state = self.__dict__.copy()
        state['parenthesis_memory'] = self.parenthesis_memory
        return state

    def __setstate__(self, state):
        '''set state for pickling. pickle default ignores class attr (parenthesis_memory) but we need it.'''
        mem = state.pop('parenthesis_memory')
        if hasattr(super(), '__setstate__'):
            super().__setstate__(state)
        else:
            self.__dict__.update(state)
        self.parenthesis_memory.update(mem)
        # important: ensure that self.parenthesis_memory is the same as the class attr;
        #   i.e., didn't accidentally assign self.parenthesis_memory to a different object.
        #   this matters because some methods might assume self.parenthesis_memory is the class attr.
        #   (first attempt at __setstate__ code had this bug!)
        assert self.parenthesis_memory is type(self).parenthesis_memory, "setstate messed up! coding error."



"""
File Purpose: loading fluid-dependent eppic.i inputs,
from values in an EppicInstabilityCalculator.
"""
import numpy as np
import xarray as xr

from ....quantities import QuantityLoader
from ....tools import simple_property
from ....defaults import DEFAULTS


class EppicDistInputsLoader(QuantityLoader):
    '''loads fluid-dependent eppic.i inputs, from values in an EppicInstabilityCalculator.'''

    # # # DIST_INPUTS -- "DISPATCH" # # #

    # {varname: name for eppic.i file.}
    # with {x} indicating "replace with component (x, y, or z)",
    #  and {N} indicating "replace with fluid index".
    # name can be dict --> expect varname to expand into vars upon loading.
    DIST_INPUTS = {
        # specie setup & "cost" for this specie.
        'init_dist': 'init_dist{N}',
        'safe_nptotcelld': 'nptotcelld{N}',
        'part_pad': 'part_pad{N}',
        'safe_pow2_subcycle': 'subcycle{N}',
        # physical parameters
        'm': 'md{N}',
        'q': 'qd{N}',
        'n': 'n0d{N}',
        'nusn': 'coll_rate{N}',
        'coll_type': 'coll_type{N}',  # (not physical, but wanted to keep it next to coll_rate.)
        # physical parameters - probably derived / equilibrium values
        'eqperp_vtherm': 'v{x}thd{N}',
        'u_drift': 'v{x}0d{N}',
        # output-related vars
        'part_out_subcycle': 'part_out_subcycle{N}',
        'flux_out_subcycle': 'flux_out_subcycle{N}',
        'nvsqr_out_subcycle': 'nvsqr_out_subcycle{N}',
        'vdist_out_subcycle': 'vdist_out_subcycle{N}',
        'vdist_lims': {
            'vdist_min': 'pv{x}min{N}',
            'vdist_max': 'pv{x}max{N}',
        },
        'pnvx': 'pnvx{N}',
        'pnvy': 'pnvy{N}',
        'pnvz': 'pnvz{N}',
    }

    @known_var(deps=DIST_INPUTS)
    def get_dist_inputs(self):
        '''Dataset of dist-related input values which go to eppic.i, here using PlasmaCalcs datavar names.
        Named & dimensioned here with PlasmaCalcs conventions,
            e.g. uses 'vdist_min' which varies across 'component' and 'fluid' dims,
            instead of labeling pvxmin0, pvymin0, pvzmin0, pvxmin1, ..., pvzminN like in eppic.i.
        Result has keys:
            (physical, direct inputs)
                'n', 'm', 'q', 'nusn',
            (physical, derived from other params)
                'u_drift', 'eqperp_vtherm',
            (numerical, direct inputs, may be input via 'dist_vals' during __init__)
                'part_pad', 'coll_type', 'pnvx', 'pnvy', 'pnvz',
                'vdist_out_subcycle', 'part_out_subcycle', 'flux_out_subcycle', 'nvsqr_out_subcycle',
            (numerical, derived from other params)
                'safe_nptotcelld', 'safe_pow2_subcycle', 'vdist_min', 'vdist_max',

        if self.dspace_mode doesn't start with 'safe_', use 'safe_{mode}' instead.
            (this affects, e.g., safe_pow2_subcycle)
        '''
        with self.using(_cached_safe_pow2_subcycle=self('safe_pow2_subcycle')):  # [EFF] caching improves efficiency.
            return self(list(self.DIST_INPUTS), dspace_mode=self.safe_dspace_mode)


    # # # VELOCITY OUTPUTS # # #

    cls_behavior_attrs.register('vdist_nsigma', default=DEFAULTS.EPPIC.VDIST_NSIGMA)
    vdist_nsigma = simple_property('_vdist_nsigma', setdefault=lambda: DEFAULTS.EPPIC.VDIST_NSIGMA,
        doc=f'''number of standard deviations (actually: vtherm widths) to output in vdist.
        vdist pvxmin & max will be centered on u_drift, with width vdist_nsigma * vtherm.
        default: DEFAULTS.EPPIC.VDIST_NSIGMA (default: {DEFAULTS.EPPIC.VDIST_NSIGMA})''')

    @known_var(deps=['u_drift', 'eqperp_vtherm'])
    def get_vdist_lims(self):
        '''Dataset of vdist_min and vdist_max for each species and each vector component (x,y,z).
        Centered on u_drift, with width self.vdist_nsigma * self('eqperp_vtherm').
        '''
        center = self('u_drift')
        width = self('eqperp_vtherm')
        vmin = center - self.vdist_nsigma * width
        vmax = center + self.vdist_nsigma * width
        return xr.Dataset({'vdist_min': vmin, 'vdist_max': vmax})


    # # # SUBCYCLING AND NUMBER OF PARTICLES # # #

    cls_behavior_attrs.register('subcycle_safety', default=DEFAULTS.EPPIC.SUBCYCLE_SAFETY)
    subcycle_safety = simple_property('_subcycle_safety', setdefault=lambda: DEFAULTS.EPPIC.SUBCYCLE_SAFETY,
        doc='''safety factor for self('safe_pow2_subcycle'). Larger is safer. None <--> 1.
        when making eppic input deck, will use subcycle = largest 2^N <= (best possible subcycling / safety)''')
    # def get_safe_pow2_subcycle... already defined in TimescalesLoader.

    cls_behavior_attrs.register('npd_mul_cpu_cost', default=DEFAULTS.EPPIC.NPD_MUL_CPU_COST)
    npd_mul_cpu_cost = simple_property('_npd_mul_cpu_cost', setdefault=lambda: DEFAULTS.EPPIC.NPD_MUL_CPU_COST,
        doc=f'''"target cpu cost" of each subcycled species, when getting self('safe_npd_mul')
        the idea is to use npd(subcycled) = npd(unsubcycled) * npd_mul_cpu_cost * subcycling
        e.g. if dist 1 has subcycling = 32, and npd_mul_cpu_cost = 0.1,
            and dist 0 has subcycling = 1, npd=1000, then dist 1 should target npd = 3200.
        (might be lower or higher due to rounding; see npd_mul_max and npd_mul_increment.)
        default: DEFAULTS.EPPIC.NPD_MUL_CPU_COST (default: {DEFAULTS.EPPIC.NPD_MUL_CPU_COST})''')

    cls_behavior_attrs.register('npd_mul_max', default=DEFAULTS.EPPIC.NPD_MUL_MAX)
    npd_mul_max = simple_property('_npd_mul_max', setdefault=lambda: DEFAULTS.EPPIC.NPD_MUL_MAX,
        doc=f'''maximum value for self('safe_npd_mul'). None --> no maximum.
        e.g. if self.npd_mul_cpu_cost = 0.1, self.npd_mul_max = 5,
            if dist 1 has subcycling = 256, use npd_mul of 5 instead of 25.6.
        default: DEFAULTS.EPPIC.NPD_MUL_MAX (default: {DEFAULTS.EPPIC.NPD_MUL_MAX})''')

    @known_var(deps=['safe_pow2_subcycle'])
    def get_safe_npd_mul(self):
        '''npd multiplier for each fluid. Aim for "target cpu cost" for subcycled fluids.
        Target cpu cost is determined by self.npd_mul_cpu_cost * (cost of unsubcycled fluid).
        The idea is to use npd(subcycled) = npd(unsubcycled) * npd_mul_cpu_cost * subcycling.
            --> safe_npd_mul = npd_mul_cpu_cost * safe_pow2_subcycle.
        However, if this would be larger than self.npd_mul_max, use npd_mul_max instead.
        Also, if this would be less than 1, use 1 instead.
        '''
        npd_mul_cpu_cost = self.npd_mul_cpu_cost
        npd_mul_max = self.npd_mul_max
        result = npd_mul_cpu_cost * self('safe_pow2_subcycle')
        if self.npd_mul_max is not None:
            result = np.minimum(result, npd_mul_max)
        result = np.maximum(result, 1)
        return result


    # # # N PIC PER CELL # # #

    @known_var
    def get_nptotcelld0(self):
        '''nptotcelld for the unsubcycled distribution (not necessarily dist 0).
        default: DEFAULTS.EPPIC.NPTOTCELLD0
        '''
        return self.load_direct('nptotcelld0')

    cls_behavior_attrs.register('npd_rounding', default=DEFAULTS.EPPIC.NPD_ROUNDING)
    npd_rounding = simple_property('_npd_rounding', setdefault=lambda: DEFAULTS.EPPIC.NPD_ROUNDING,
        doc=f'''rounding to use for npd values. None --> no rounding (still round to nearest int)
        e.g. use 10 if you want all npd choices to be rounded to the nearest 10.
        default: DEFAULTS.EPPIC.NPD_ROUNDING (default: {DEFAULTS.EPPIC.NPD_ROUNDING})''')

    @known_var(deps=['nptotcelld0', 'safe_npd_mul'])
    def get_safe_nptotcelld(self):
        '''nptotcelld to use for each fluid.
        Usually, safe_nptotcelld = nptotcelld(the unsubcycled dist) * self('safe_npd_mul').
            internally, uses nptotcelld(unsubcycled) = self.nptotcelld0.
        Rounds to nearest int, or nearest multiple of self.npd_rounding, if non-None.
        '''
        result = self('nptotcelld0') * self('safe_npd_mul')
        if self.npd_rounding is None:
            result = np.round(result)
        else:
            result = np.round(result / self.npd_rounding) * self.npd_rounding
        return result.astype('int')



"""
Package Purpose: optional add-ons for PlasmaCalculators

[TODO] custom gettable vars (defined outside of PlasmaCalcs)?
[TODO] "load all addons" function; don't load until function called.
        this way, user could edit DEFAULTS before loading addons.
"""
from . import addon_tools
# load all addon modules (allowing them to register_addon_loader if appropriate)
from . import tfbi
from . import instability_tools

# import specific things from addon modules
from .instability_tools import itAccessor
from .addon_tools import ADDON_LOADERS

# create AddonLoader object with all loaded addon loaders
class AddonLoader(*ADDON_LOADERS):
    '''loader for all (successfully) imported addons.'''
    pass



"""
File Purpose: QuantityLoader.help()

defines QuantityLoader.help() which helps with loading quantities,
particularly with learning which quantities are available to be loaded.
"""
import builtins
import textwrap

from .quantity_loader import QuantityLoader
from .quantity_tools import Pattern
from ..defaults import DEFAULTS
from ..errors import FormulaMissingError, InputError
from ..tools import (
    format_docstring, indent_paramdocs,
    help_str, _help_str_paramdocs,
    UNSET,
    Binding,
)
binding = Binding(locals())


with binding.to(QuantityLoader):
    QuantityLoader.__doc__ += '''\n    For help with loading quantities, see QuantityLoader.help()\n'''

    _help_quants_str_paramdocs = {
        **indent_paramdocs(_help_str_paramdocs),
        'qstr': '''None or str
            None --> tells info about this class & how to use this function.
                    in particular, tells that quants are stored cls.KNOWN_VARS and cls.KNOWN_PATTERNS,
                    and describes behavior of calling help with a string.
            str --> return str for help with all quants related to str.
                    use empty str to get help for all quants.''',
        'only': '''None or str
            None --> get help with all quantities related to qstr.
            'VARS' or 'vars' --> only get help with KNOWN_VARS.
            'PATTERNS' or 'patterns' --> only get help with KNOWN_PATTERNS.
            'TREE' or 'tree' --> only get help with quantities in cls.cls_var_tree(str).
            if provided when qstr is None, treat qstr as '' instead.''',
        'dense': '''bool
            Whether to reduce whitespace in result.
            E.g. True --> no newlines between functions. False --> one newline between functions.''',
        'modules': '''bool
            Whether to include modules in result.
            If True, result will be grouped into sections with modules written at top.''',
        'tree': '''None or bool
            How much help to give for quantities in cls.cls_var_tree(qstr).
            False --> don't even check cls.cls_var_tree(qstr).
            True --> help for all quantities in cls.cls_var_tree.
            None --> help for quantities in cls.cls_var_tree(qstr).flat_branches_until_vars()
                    i.e. patterns & vars in tree but ignore any nodes with LoadableVar ancestors.
                    e.g. qstr='mean_mod_beta' --> help with 'mean_(.+)', 'mod_(.+)', and 'beta',
                        but no help with dependencies of 'beta' ('q', 'mod_B', 'm').''',
        'print': '''bool
            whether to print the result. If False, return the result instead of printing.''',
    }
    QuantityLoader._help_quants_str_paramdocs = _help_quants_str_paramdocs

    @binding.bind(methodtype=classmethod)
    def _help_matches(cls, qstr, k, v):
        '''returns whether qstr matches k or v, and thus should be displayed during self.help(qstr).

        qstr: str
            the str to match; from self.help(qstr)
        k: varname
            the varname to test for matches.
            key from self.KNOWN_VARS.keys(), or key.str from self.KNOWN_PATTERNS.keys().
        v: LoadableQuantity
            the LoadableQuantity to test for matches.
            value from self.KNOWN_VARS.values() or self.KNOWN_PATTERNS.values().

        matches if any of these are true:
            qstr == ''
            qstr in k.split('_')         # size limitation and split('_') because, e.g. during help('n'),
            len(qstr)>=3 and qstr in k   #    want vars related to number density, not all vars with the letter 'n'.
            qstr in module.split('.')   (where, module == v.get_f_module(cls))
            '.' in qstr and qstr in module
            len(qstr)>=3 and qstr in value from module.split('.')
            len(qstr)>=3 and qstr in v.fname
            re.fullmatch(k, qstr)  # if k is a Pattern
        otherwise, does not match.
        '''
        if qstr == '': return True
        if qstr in k.split('_'): return True
        if len(qstr)>=3 and qstr in k: return True
        module = v.get_f_module(cls)
        if qstr in module.split('.'): return True
        if '.' in qstr and qstr in module: return True
        if len(qstr)>=3 and any(qstr in val for val in module.split('.')): return True
        if len(qstr)>=3 and qstr in v.fname: return True
        if isinstance(k, Pattern) and k.fullmatch(qstr): return True
        return False

    @binding.bind(methodtype=classmethod)
    @format_docstring(**_help_quants_str_paramdocs)
    def help_quants_str(cls, qstr=None, only=None, *, tree=None, modules=True, signature=False, doc=True, dense=False,
                        _instance=None):
        '''returns str for help with quants.

        qstr: {qstr}
        only: {only}
        tree: {tree}
        modules: {modules}
        signature: {signature}
        doc: {doc}
        dense: {dense}
        _instance: None or QuantityLoader instance
            if provided, use _instance.match_var_tree() instead of cls.cls_var_tree().
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        TAB = ' '*4
        if only is not None:
            only = only.upper()  # case-insensitive
            if only not in ('VARS', 'PATTERNS', 'TREE'):
                raise InputError(f'only={only!r} not recognized; expected "VARS", "PATTERNS", or "TREE".')
            if qstr is None:
                qstr = ''   # treat qstr as '' when qstr=None but only is not None.
        if qstr is None:
            helpstr = f'''\
            - obj('qname') or obj.get('qname') --> get the value of quantity qname.

            - obj.help(str) shows help for all related quantities:
                - obj.help('') --> help with all available quantities.
                - obj.help(str) --> get help with all quantities related to that str:
                    - quantities containing str
                    - quantities from module with name matching str

                - obj.help(str, 'VARS') --> help with only KNOWN_VARS.
                - obj.help(str, 'PATTERNS') --> help with only KNOWN_PATTERNS.
                - obj.help(str, 'TREE') --> only show help for quantities in match_var_tree(str).
                - obj.help(str, ..., tree=True) --> include help for ALL quantities in match_var_tree(str),
                    instead of ignoring tree nodes which have any LoadableVar ancestors.

                - obj.help(str, ..., modules=True) --> include module info in the output;
                    this helps with understanding where the code for each quantity is defined.

            - {cls.__name__}.cls_help(str) works similarly, but:
                - doesn't require creating an instance of {cls.__name__} to get help.
                - will fail for quants which depend on present values of a {cls.__name__} instance.
                    e.g. {cls.__name__}.cls_help('nusj') fails if collision type depends on obj.fluid.

            - {cls.__name__}.KNOWN_VARS and .KNOWN_PATTERNS store all available quantities.
            - obj.behavior tells all attributes of {cls.__name__} which may affect results.
            - obj.help_call_options() tells all optional kw during obj('qname', **kw).

            - obj.match_var_tree('qname') shows how quantity qname will actually be loaded,
                including any quantities which qname depends on, and any which those depend on, etc.'''
            return textwrap.dedent(helpstr)
        # else:  qstr is a str. find matches
        tree_flag = tree
        if (only in (None, 'TREE')) and (tree_flag or (tree_flag is None)):
            try:
                if _instance is None:
                    tree = cls.cls_var_tree(qstr)
                else:
                    tree = _instance.match_var_tree(qstr)
            except FormulaMissingError:
                tree_loadables = []
            else:
                if tree_flag:
                    tree_loadables = [leaf.obj.loadable for leaf in tree.flat(include_self=True)]
                else:
                    tree_loadables = [leaf.obj.loadable for leaf in tree.flat_branches_until_vars(include_self=True)]
        else:
            tree_loadables = []

        vars_match = {}
        if only in (None, 'VARS'):  # add any VARS matches
            vars_match = {k:v for k,v in cls.KNOWN_VARS.items() if cls._help_matches(qstr, k, v)}
        if only in (None, 'TREE'):  # add any TREE matches
            vars_match = {**vars_match, **{k:v for k,v in cls.KNOWN_VARS.items() if v in tree_loadables}}

        pats_match = {}
        if only in (None, 'PATTERNS'):  # add any PATTERNS matches
            pats_match = {k:v for k,v in cls.KNOWN_PATTERNS.items() if cls._help_matches(qstr, k.str, v)}
        if only in (None, 'TREE'):  # add any TREE matches
            pats_match = {**pats_match, **{k:v for k,v in cls.KNOWN_PATTERNS.items() if v in tree_loadables}}

        if len(vars_match) == 0 and len(pats_match) == 0:
            return f'No quantities containing string: {qstr!r}'
        multiple_vars_match = (len(vars_match) > 1)
        multiple_pats_match = (len(pats_match) > 1)
        showing_vars_and_pats = (len(vars_match) > 0) and (len(pats_match) > 0)

        # include summary of all matches at top
        result = ''
        if showing_vars_and_pats or multiple_vars_match:
            result += f'Showing help for related KNOWN_VARS: {list(vars_match.keys())}\n'
            if not dense: result += '\n'
        if showing_vars_and_pats or multiple_pats_match:
            result += f'Showing help for related KNOWN_PATTERNS: {[key.str for key in pats_match.keys()]}\n'
            if not dense: result += '\n'

        # group vars & pattern matches by module
        if modules:
            vars_match_by_mod = dict()
            pats_match_by_mod = dict()
            for k, v in vars_match.items():
                mod = v.get_f_module(cls)
                vars_match_by_mod.setdefault(mod, dict())[k] = v
            for k, v in pats_match.items():
                mod = v.get_f_module(cls)
                pats_match_by_mod.setdefault(mod, dict())[k] = v
        else:
            vars_match_by_mod = {0: vars_match}
            pats_match_by_mod = {0: pats_match}

        # KNOWN_VARS
        if len(vars_match) > 0:
            if showing_vars_and_pats: result += f'\nKNOWN_VARS:\n'
            result_vars = ''
            for mod, vars_match in vars_match_by_mod.items():
                if modules:
                    if not dense: result_vars += '\n'
                    result_vars += f'in module {mod}:\n'
                result_mod = ''
                for k, v in vars_match.items():
                    if signature:   # put "key  : signature" as first line.
                        result_mod += f'{k!r:7s}: '
                    elif doc:       # put "key:" as first line. (don't put doc info on first line.)
                        result_mod += f'{k!r}:\n'
                    if signature or doc:
                        result_mod += help_str(v.get_f(cls), module=False, signature=signature, doc=doc)
                        result_mod += '\n'
                        if not dense and not modules: result_mod += '\n'
                    else:  # no signature & no doc. Just a list of vars.
                        result_mod += f'{k!r}, '
                if modules:
                    result_mod = textwrap.indent(result_mod, TAB)
                    if (not signature) and (not doc): result_mod += '\n'
                result_vars += result_mod
            if showing_vars_and_pats: result_vars = textwrap.indent(result_vars, TAB)
            result += result_vars
            if showing_vars_and_pats and modules and not dense: result += '\n'

        # KNOWN_PATTERNS
        if len(pats_match) > 0:
            if showing_vars_and_pats: result += f'\nKNOWN_PATTERNS:\n'
            result_pats = ''
            for mod, pats_match in pats_match_by_mod.items():
                if modules:
                    if not dense: result_pats += '\n'
                    result_pats += f'in module {mod}:\n'
                result_mod = ''
                for k, v in pats_match.items():
                    if signature:   # put "key  : signature" as first line.
                        result_mod += f'{k.str!r:25s}: '
                    elif doc:       # put "key:" as first line. (don't put doc info on first line.)
                        result_mod += f'{k.str!r}:\n'
                    if signature or doc:
                        result_mod += help_str(v.get_f(cls), module=False, signature=signature, doc=doc)
                        result_mod += '\n'
                        if not dense and not modules: result_mod += '\n'
                    else:   # no signature & no doc. Just a list of patterns.
                        result_mod += f'{k.str!r}, '
                if modules:
                    result_mod = textwrap.indent(result_mod, TAB)
                    if (not signature) and (not doc): result_mod += '\n'
                result_pats += result_mod
            if showing_vars_and_pats: result_pats = textwrap.indent(result_pats, TAB)
            result += result_pats

        result = result.strip('\n')  # remove leading / trailing newlines
        return result

    @binding.bind(methodtype=classmethod)
    def help_str(cls, qstr=None, only=None, **kw):
        '''returns cls.help_quants_str(qstr=qstr, only=only, **kw).
        cls.help() calls help_str.
        subclasses might overwrite help_str, but probably won't touch help_quants_str.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        return cls.help_quants_str(qstr=qstr, only=only, **kw)
    
    @binding.bind(methodtype=classmethod)
    @format_docstring(**_help_quants_str_paramdocs)
    def cls_help(cls, qstr=None, only=None, *, tree=None, modules=False, signature=False,
                 doc=True, dense=False, print=True, **kw):
        '''prints str for help with quants. Fails for any quants which depend on present values of a cls instance.

        qstr: {qstr}
        only: {only}
        tree: {tree}
        modules: {modules}
        signature: {signature}
        doc: {doc}
        dense: {dense}
        print: {print}
        '''
        helpstr = cls.help_str(qstr=qstr, only=only, tree=tree,
                                modules=modules, signature=signature, doc=doc, dense=dense,
                                **kw)
        if print:
            builtins.print(helpstr)
        else:
            return helpstr

    @binding
    @format_docstring(**_help_quants_str_paramdocs)
    def help(self=UNSET, qstr=None, only=None, *, tree=None, modules=False, signature=False,
             doc=True, dense=False, print=True):
        '''prints str for help with quants.

        qstr: {qstr}
        only: {only}
        tree: {tree}
        modules: {modules}
        signature: {signature}
        doc: {doc}
        dense: {dense}
        '''
        if not isinstance(self, QuantityLoader):
            # user tried calling help as a classmethod. Raise helpful error message about it.
            errmsg = (f'Expected QuantityLoader self, got type(self)={type(self)}.\n'
                      'This might occur if you called cls.help, instead of obj.help.\n'
                      'Use cls.cls_help() instead of cls.help(), to try getting help from class.\n'
                      '  It should succeed for most vars, but fail for vars whose tree depends on present values.')
            raise InputError(errmsg)
        helpstr = self.help_str(qstr=qstr, only=only, tree=tree,
                                modules=modules, signature=signature, doc=doc, dense=dense,
                                _instance=self)
        if print:
            builtins.print(helpstr)
        else:
            return helpstr

    @binding
    def help_call_options(self, search=None):
        '''prints help for kw_call_options.
        if search is provided, only print help for keys containing search.
        '''
        # [TODO] option to return all this stuff as a string, instead of printing
        # [TODO] option to see docs for only some of the options, instead of all at once
        print('Showing self.help_call_options(): docs for keys from self.kw_call_options()')
        if search is not None: print(f'Only showing keys containing search={search!r}')
        print('---------------------------------------------------------------------------')
        for key in self.kw_call_options():
            if search is not None and search not in key:
                continue
            s = getattr(type(self), key, None)
            if s is None:
                print(f"'{key}': [no help found; this key is not an attribute of type(self)]")
            else:
                print(f"'{key}': {help_str(s, module=False).lstrip()}")
            print()



"""
File Purpose: DimPoint and DimRegion
"""
import xarray as xr

from ...errors import DimensionValueError
from ...tools import (
    repr_simple,
    UNSET,
    is_iterable_dim,
    xarray_assign,
)

''' --------------------- DimRegion and DimPoint --------------------- '''

class DimRegion(dict):
    '''dict of values representing some region of dimension space (possibly a single point).'''
    # # # DISPLAY # # #
    def __repr__(self):
        '''like dict repr but use repr_simple(val) instead of repr(val) for values.'''
        contents = ', '.join(f'{key}={repr_simple(val)}' for key, val in self.items())
        return f'{type(self).__name__}({contents})'

    def assign_to(self, array, **kw_xarray_assign):
        '''array.assign_coords(self). Internally, use PlasmaCalcs xarray_assign function.
        xarray_assign also allows to provide attrs and/or name to assign those as well.
        Also convert array to xr.DataArray if it isn't one already.

        Probably will fail if any values are iterable dimensions e.g. DimensionValueList instead of DimensionValue.
        '''
        if not isinstance(array, xr.DataArray):
            array = xr.DataArray(array)
        return xarray_assign(array, self, **kw_xarray_assign)

class DimPoint(DimRegion):
    '''dict of values representing a single point in dimension-space.
    When setting elements of this dict, ensures not is_iterable_dim(value)

    Does not need to include all dimensions, but all contents must be single-valued dimensions.
    E.g. {'fluid': 0, 'snap': 1} is fine, even if 'jfluid' and 'component' are also dimensions.
    However, {'fluid': 0, 'component':[0,1,2]} is not, because component is multi-valued.
    '''
    def __init__(self, *args_dict, **kw_dict):
        super().__init__(*args_dict, **kw_dict)
        # dict.__init__ doesn't actually utilize __setitem__; need to check values here too.
        for key, value in self.items():
            self._check_value(value, key=key)

    def __setitem__(self, key, value):
        self._check_value(value, key=key)
        super().__setitem__(key, value)

    def _check_value(self, value, *, key=UNSET):
        '''raise DimensionValueError if value is iterable.
        if raising error, include key in error message if provided.
        '''
        if is_iterable_dim(value):
            if key is UNSET:
                errmsg = f"is_iterable_dim(value): {value!r}."
            else:
                errmsg = f"for {key!r}, is_iterable_dim(value): {value!r}."
            raise DimensionValueError(errmsg)


"""
File Purpose: testing tools from tools.xarray_tools.xarray_coords
"""
import pytest
import numpy as np
import xarray as xr

import PlasmaCalcs as pc


def test_xarray_nondim_coords():
    xx = pc.xrrange(5, name='x')
    yy = pc.xrrange(3, 'y')
    arr = xx * yy
    assert len(arr.pc.nondim_coords())==0
    arr = arr.assign_coords(z=7, label='hi')
    ndc = arr.pc.nondim_coords()
    assert ndc['z'] == 7
    assert ndc['label'] == 'hi'
    arr0 = arr.isel(x=0)
    ndc0 = arr0.pc.nondim_coords()
    assert ndc0['x'] == 0
    assert ndc0['z'] == 7
    assert ndc0['label'] == 'hi'
    arr1 = arr.isel(x=0, drop=True)
    assert 'x' not in arr1.pc.nondim_coords()
    # dataset
    ds = xr.Dataset({'xx': xx, 'yy': yy})
    ds = ds.assign_coords(z=7, label='hi')
    ndc = ds.pc.nondim_coords()
    assert ndc['z'] == 7
    assert ndc['label'] == 'hi'
    ds0 = ds.isel(x=0)
    ndc0 = ds0.pc.nondim_coords()
    assert ndc0['x'] == 0
    assert ndc0['z'] == 7
    assert ndc0['label'] == 'hi'

def test_dims_coords():
    xx = pc.xrrange(5, 'x')
    assert xx.pc.dims_coords() == {'x': ['x']}
    yy = pc.xrrange(3, 'y')
    arr = xx * yy
    assert arr.pc.dims_coords() == {'x': ['x'], 'y': ['y']}
    arr = arr.assign_coords(z=xx + yy, testscalar=7)
    dc = arr.pc.dims_coords()
    assert set(dc.keys()) == set(['x', 'y', ()])
    dc_set_vals = {k: set(v) for k,v in dc.items()}
    assert dc_set_vals == {'x': set(['x', 'z']), 'y': set(['y', 'z']), (): set(['testscalar'])}
    yy2 = pc.xrrange(3, 'y').assign_coords(height=lambda arr: arr['y']+10)
    dc = yy2.pc.dims_coords()
    assert set(dc.keys()) == set(['y'])
    assert set(dc['y']) == set(['y', 'height'])

def test_assign_self_as_coord():
    arr = (pc.xrrange(5, 'x') * pc.xrrange(3, 'y')).rename('z')
    arr0 = arr
    arr = arr.pc.assign_self_as_coord()
    assert np.all(arr == arr0)  # assigning new coord doesn't change array values
    assert np.all(arr['z'] == arr)

def test_fill_coords():
    arr = xr.DataArray(np.arange(10)*10, dims=['x'])
    assert 'x' not in arr.coords
    filled = arr.pc.fill_coords()
    assert np.all(filled['x'] == np.arange(10))
    assert 'x' in filled.coords
    arr2 = xr.DataArray(np.arange(24).reshape(2,3,4), dims=['x', 'y', 'z'], coords={'y': [5, 7, 100]})
    filled = arr2.pc.fill_coords()
    assert all(x in filled.coords for x in ['x', 'y', 'z'])
    assert np.all(filled['x'] == np.arange(2))  # existing coords untouched
    assert np.all(filled['y'] == [5, 7, 100])
    assert np.all(filled['z'] == np.arange(4))
    filled = arr2.pc.fill_coords(dim='z')
    assert 'x' not in filled.coords
    assert 'z' in filled.coords
    filled = arr2.pc.fill_coords(dim='y')
    assert filled.identical(arr2)  # no change because y already has coords.

def test_index_coords():
    arr = pc.xrrange(5, 'x') * pc.xrrange(3, 'y', coords={'y': ['a', 'b', 'c']})
    arr = arr.rename('z').pc.assign_self_as_coord()
    assert arr.pc.index_coords().identical(pc.xarray_index_coords(arr))
    ixc = arr.pc.index_coords()
    assert set(ixc.dims) == {'x', 'y'}
    assert set(ixc.coords) == {'x', 'x_index', 'y', 'y_index', 'z', 'z_index'}
    assert all(np.all(ixc[x] == arr[x]) for x in ('x', 'y', 'z'))  # original coords unchanged
    assert np.all(ixc['x_index'] == np.arange(5))
    assert np.all(ixc['y_index'] == np.arange(3))
    for i in range(5):
        for j in range(3):
            assert ixc['z_index'].isel(x=i, y=j).item() == (i, j)
    ixc = arr.pc.index_coords(drop=True)
    assert set(ixc.dims) == {'x', 'y'}
    assert set(ixc.coords) == {'x_index', 'y_index', 'z_index'}
    with pytest.raises(pc.DimensionalityError):
        ixc = arr.pc.index_coords(promote=True)  # can't promote z to dim because z is 2D.
    ixc = arr.pc.index_coords(('x', 'y'))
    assert set(ixc.dims) == {'x', 'y'}
    assert set(ixc.coords) == {'x', 'x_index', 'y', 'y_index', 'z'}
    ixc = arr.pc.index_coords(('x', 'y'), drop=True)
    assert set(ixc.dims) == {'x', 'y'}
    assert set(ixc.coords) == {'x_index', 'y_index', 'z'}
    ixc = arr.pc.index_coords(('x', 'y'), promote=True)
    assert set(ixc.dims) == {'x_index', 'y_index'}
    assert set(ixc.coords) == {'x', 'x_index', 'y', 'y_index', 'z'}
    ixc = arr.pc.index_coords(('x', 'y'), drop=True, promote=True)
    assert set(ixc.dims) == {'x_index', 'y_index'}
    assert set(ixc.coords) == {'x_index', 'y_index', 'z'}
    with pytest.raises(pc.InputError):
        idx = arr.pc.index_coords('x', 'z')  # <-- some new names (i.e., 'z') already exist in array.coords.

def test_manip_coords():
    '''tests scale_coords, shift_coords, and log_coords.'''
    # non-comprehensive (not getting all edge cases) but it's something at least.
    ## syntaxes
    arr = (pc.xrrange(5, 'x') * pc.xrrange(3, 'y')).rename('z')
    assert arr.pc.shift_coords({'x': 2}).identical(arr.pc.shift_coords(x=2))
    assert arr.pc.shift_coords({'x': 2}).identical(pc.xarray_shift_coords(arr, {'x': 2}))
    assert arr.pc.scale_coords({'x': 2}).identical(arr.pc.scale_coords(x=2))
    assert arr.pc.scale_coords({'x': 2}).identical(pc.xarray_scale_coords(arr, {'x': 2}))
    xx = pc.xrrange(5, 'x').pc.shift_coords({'x':1})
    yy = pc.xrrange(3, 'y', coords={'y': [1, 10, np.e]})
    arr = (xx * yy).rename('z')
    assert arr.pc.log_coords().identical(pc.xarray_log_coords(arr))
    assert arr.pc.log_coords().identical(arr.pc.log_coords(drop=True, promote=False))  # <- check defaults
    ## scale coords
    arr = (pc.xrrange(5, 'x') * pc.xrrange(3, 'y')).rename('z')
    x2 = arr.pc.scale_coords({'x': 2})
    assert np.all(x2['x'] == np.arange(5)*2)
    assert x2.identical(arr.pc.scale_coords(x=2))  # alternate syntax
    x2_y10 = arr.pc.scale_coords({'x': 2, 'y': 10})
    assert np.all(x2_y10['x'] == np.arange(5)*2)
    assert np.all(x2_y10['y'] == np.arange(3)*10)

    ## shift coords
    x2 = arr.pc.shift_coords({'x': 2})
    assert np.all(x2['x'] == np.arange(5)+2)
    xm2_y10 = arr.pc.shift_coords({'x': -2, 'y': 10})
    assert np.all(xm2_y10['x'] == np.arange(5)-2)
    assert np.all(xm2_y10['y'] == np.arange(3)+10)

    ## log coords
    xx = pc.xrrange(5, 'x').pc.shift_coords({'x':1})
    yy = pc.xrrange(3, 'y', coords={'y': [1, 10, np.e]})
    arr = (xx * yy).rename('z')
    arr = arr.assign_coords({'testscalar': 1000})
    log = arr.pc.log_coords(drop=True, promote=False)  # <-- this is the default behavior, as checked above.
    assert set(log.dims) == {'x', 'y'}
    assert set(log.coords) == {'log_x', 'log_y', 'log_testscalar'}
    log = arr.pc.log_coords(drop=True, promote=True)
    assert set(log.dims) == {'log_x', 'log_y', 'log_testscalar'}
    assert set(log.coords) == {'log_x', 'log_y', 'log_testscalar'}
    log = arr.pc.log_coords(drop=False, promote=True)
    assert set(log.dims) == {'log_x', 'log_y', 'log_testscalar'}
    assert set(log.coords) == {'log_x', 'log_y', 'log_testscalar', 'x', 'y', 'testscalar'}
    log = arr.pc.log_coords(drop=False, promote=False)
    assert set(log.dims) == {'x', 'y'}
    assert set(log.coords) == {'log_x', 'log_y', 'log_testscalar', 'x', 'y', 'testscalar'}
    assert np.all(log['log_x'] == np.log10(xx['x']))
    assert np.all(log['log_y'] == np.log10(yy['y']))
    assert log['log_y'].isel(y=0) == 0
    assert log['log_y'].isel(y=1) == 1
    # check some more values:
    logyrep = arr.pc.log_coords(['y', 'testscalar'], drop=True, promote=False)
    assert logyrep['log_testscalar'] == 3
    assert logyrep['log_y'].isel(y=1) == 1
    logbasee = arr.pc.log_coords(base='e')
    assert np.allclose(logbasee['log_y'], np.log(yy['y']))
    assert logbasee.attrs['log_base'] == 'e'
    assert logbasee['log_y'].isel(y=2) == 1
    # check what happens if newname is altered:
    log = arr.pc.log_coords(newname='customlog_{coord}')
    assert set(log.dims) == {'x', 'y'}
    assert set(log.coords) == {'customlog_x', 'customlog_y', 'customlog_testscalar'}
    log = arr.pc.log_coords(newname='{coord}')
    assert set(log.dims) == {'x', 'y'}
    assert set(log.coords) == {'x', 'y', 'testscalar'}
    assert np.all(log['x'] == np.log10(xx['x']))
    with pytest.raises(pc.InputError):
        arr.pc.log_coords(newname='same_for_all')  # <-- multiple coords would have same new name --> crash.


# [TODO] tests for: xarray_is_sorted, get_dx_along, differentiate



"""
Package Purpose: InstabilityCalculator specialized in computing EPPIC values.
Input is an xarray.Dataset.
Knows how to write input deck files for eppic, based on values.
"""

from .eppic_instability_calculator import EppicInstabilityCalculator
from .eppic_tfbi_calculator import EppicTfbiCalculator, EppicCalculatorWithTfbi

from .eppic_dist_inputs_loader import EppicDistInputsLoader
from .eppic_glob_inputs_loader import EppicGlobInputsLoader
from .eppic_safety_info_loader import EppicSafetyInfoLoader
from .eppici_maker import (
    EppiciMaker,
    eppici_dict, eppici,
)


# import submodule into this namespace so it can be accessed from elsewhere if needed.
# (necessary because it has the same name as this subpackage: eppic_instability_calculator)
from . import eppic_instability_calculator as _eppic_instability_calculator_py
